# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os 
import re
import json
import collections
from itertools import islice, chain
from collections import OrderedDict, defaultdict
from typing import Optional, Callable, Tuple, Dict, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.hub import load_state_dict_from_url


def load_checkpoint(model, checkpoint_path):
    if checkpoint_path and os.path.isfile(checkpoint_path):
        checkpoint = torch.load(checkpoint_path, map_location='cpu')
    else:
        checkpoint = load_state_dict_from_url(checkpoint_path, map_location='cpu')

    
    orig_state_dict = checkpoint['model']
    new_state_dict = {}
    for key, item in orig_state_dict.items():

        
        if key.startswith('module'):
            key = '.'.join(key.split('.')[1:])
        
        # TODO: better ways
        if key.startswith('fc') or key.startswith('classifier') or key.startswith('mlp') or key.startswith('head'):
            continue
            
        # check vit and interpolate
        # if isinstance(model, VisionTransformer) and 'patch_emb'

        if key == 'pos_embed':
            posemb_new = model.pos_embed.data
            posemb = item
            item = resize_pos_embed_vit(posemb, posemb_new)

        new_state_dict[key] = item 
    
    match = model.load_state_dict(new_state_dict, strict=False)
    print(match)
    return model



def resize_pos_embed_vit(posemb, posemb_new, num_tokens=1, gs_new=()):
    # Rescale the grid of position embeddings when loading from state_dict. Adapted from
    # https://github.com/google-research/vision_transformer/blob/00883dd691c63a6830751563748663526e811cee/vit_jax/checkpoint.py#L224
    # _logger.info('Resized position embedding: %s to %s', posemb.shape, posemb_new.shape)
    import math
    ntok_new = posemb_new.shape[1]
    if num_tokens:
        posemb_tok, posemb_grid = posemb[:, :num_tokens], posemb[0, num_tokens:]
        ntok_new -= num_tokens
    else:
        posemb_tok, posemb_grid = posemb[:, :0], posemb[0]
    gs_old = int(math.sqrt(len(posemb_grid)))
    if not len(gs_new):  # backwards compatibility
        gs_new = [int(math.sqrt(ntok_new))] * 2
    assert len(gs_new) >= 2
    # _logger.info('Position embedding grid-size from %s to %s', [gs_old, gs_old], gs_new)
    posemb_grid = posemb_grid.reshape(1, gs_old, gs_old, -1).permute(0, 3, 1, 2)
    posemb_grid = F.interpolate(posemb_grid, size=gs_new, mode='bicubic', align_corners=False)
    posemb_grid = posemb_grid.permute(0, 2, 3, 1).reshape(1, gs_new[0] * gs_new[1], -1)
    posemb = torch.cat([posemb_tok, posemb_grid], dim=1)
    return posemb



def param_groups_weight_decay(
        model: nn.Module,
        weight_decay=1e-5,
        no_weight_decay_list=()
):
    # Ref: https://github.com/rwightman/pytorch-image-models
    no_weight_decay_list = set(no_weight_decay_list)
    decay = []
    no_decay = []
    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue

        if param.ndim <= 1 or name.endswith(".bias") or name in no_weight_decay_list:
            no_decay.append(param)
        else:
            decay.append(param)

    return [
        {'params': no_decay, 'weight_decay': 0.},
        {'params': decay, 'weight_decay': weight_decay}]


def _group(it, size):
    it = iter(it)
    return iter(lambda: tuple(islice(it, size)), ())


def _layer_map(model, layers_per_group=12, num_groups=None):
    def _in_head(n, hp):
        if not hp:
            return True
        elif isinstance(hp, (tuple, list)):
            return any([n.startswith(hpi) for hpi in hp])
        else:
            return n.startswith(hp)

    head_prefix = getattr(model, 'pretrained_cfg', {}).get('classifier', None)
    names_trunk = []
    names_head = []
    for n, _ in model.named_parameters():
        names_head.append(n) if _in_head(n, head_prefix) else names_trunk.append(n)

    # group non-head layers
    num_trunk_layers = len(names_trunk)
    if num_groups is not None:
        layers_per_group = -(num_trunk_layers // -num_groups)
    names_trunk = list(_group(names_trunk, layers_per_group))

    num_trunk_groups = len(names_trunk)
    layer_map = {n: i for i, l in enumerate(names_trunk) for n in l}
    layer_map.update({n: num_trunk_groups for n in names_head})
    return layer_map


def group_parameters(
        module: nn.Module,
        group_matcher,
        output_values=False,
        reverse=False,
):
    # Ref: https://github.com/rwightman/pytorch-image-models
    return group_with_matcher(
        module.named_parameters(), group_matcher, output_values=output_values, reverse=reverse)


def param_groups_layer_decay(
        model: nn.Module,
        lr : float=1e-3,
        weight_decay: float = 0.05,
        no_weight_decay_list: Tuple[str] = (),
        layer_decay: float = .75,
        end_layer_decay: Optional[float] = None,
):
    """
    Parameter groups for layer-wise lr decay & weight decay
    Based on BEiT: https://github.com/microsoft/unilm/blob/master/beit/optim_factory.py#L58
    # Ref: https://github.com/rwightman/pytorch-image-models
    """
    no_weight_decay_list = set(no_weight_decay_list)
    param_group_names = {}  # NOTE for debugging
    param_groups = {}

    if hasattr(model, 'group_matcher'):
        # FIXME interface needs more work
        layer_map = group_parameters(model, model.group_matcher(coarse=False), reverse=True)
    else:
        # fallback
        layer_map = _layer_map(model)
    num_layers = max(layer_map.values()) + 1
    layer_max = num_layers - 1
    layer_scales = list(layer_decay ** (layer_max - i) for i in range(num_layers))

    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue

        # no decay: all 1D parameters and model specific ones
        if param.ndim == 1 or name in no_weight_decay_list:
            g_decay = "no_decay"
            this_decay = 0.
        else:
            g_decay = "decay"
            this_decay = weight_decay

        layer_id = layer_map.get(name, layer_max)
        group_name = "layer_%d_%s" % (layer_id, g_decay)

        if group_name not in param_groups:
            this_scale = layer_scales[layer_id]
            param_group_names[group_name] = {
                "lr": this_scale * lr,
                "weight_decay": this_decay,
                "param_names": [],
            }
            param_groups[group_name] = {
                "lr": this_scale * lr,
                "weight_decay": this_decay,
                "params": [],
            }

        param_group_names[group_name]["param_names"].append(name)
        param_groups[group_name]["params"].append(param)

    # FIXME temporary output to debug new feature
    # print("parameter groups: \n%s" % json.dumps(param_group_names, indent=2))

    return list(param_groups.values())


MATCH_PREV_GROUP = (99999,)

def group_with_matcher(
        named_objects,
        group_matcher: Union[Dict, Callable],
        output_values: bool = False,
        reverse: bool = False
):
    # Ref: https://github.com/rwightman/pytorch-image-models
    if isinstance(group_matcher, dict):
        # dictionary matcher contains a dict of raw-string regex expr that must be compiled
        compiled = []
        for group_ordinal, (group_name, mspec) in enumerate(group_matcher.items()):
            if mspec is None:
                continue
            # map all matching specifications into 3-tuple (compiled re, prefix, suffix)
            if isinstance(mspec, (tuple, list)):
                # multi-entry match specifications require each sub-spec to be a 2-tuple (re, suffix)
                for sspec in mspec:
                    compiled += [(re.compile(sspec[0]), (group_ordinal,), sspec[1])]
            else:
                compiled += [(re.compile(mspec), (group_ordinal,), None)]
        group_matcher = compiled

    def _get_grouping(name):
        if isinstance(group_matcher, (list, tuple)):
            for match_fn, prefix, suffix in group_matcher:
                r = match_fn.match(name)
                if r:
                    parts = (prefix, r.groups(), suffix)
                    # map all tuple elem to int for numeric sort, filter out None entries
                    return tuple(map(float, chain.from_iterable(filter(None, parts))))
            return float('inf'),  # un-matched layers (neck, head) mapped to largest ordinal
        else:
            ord = group_matcher(name)
            if not isinstance(ord, collections.abc.Iterable):
                return ord,
            return tuple(ord)

    # map layers into groups via ordinals (ints or tuples of ints) from matcher
    grouping = defaultdict(list)
    for k, v in named_objects:
        grouping[_get_grouping(k)].append(v if output_values else k)

    # remap to integers
    layer_id_to_param = defaultdict(list)
    lid = -1
    for k in sorted(filter(lambda x: x is not None, grouping.keys())):
        if lid < 0 or k[-1] != MATCH_PREV_GROUP[0]:
            lid += 1
        layer_id_to_param[lid].extend(grouping[k])

    if reverse:
        assert not output_values, "reverse mapping only sensible for name output"
        # output reverse mapping
        param_to_layer_id = {}
        for lid, lm in layer_id_to_param.items():
            for n in lm:
                param_to_layer_id[n] = lid
        return param_to_layer_id

    return layer_id_to_param


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
import wandb
from .hook import Hook


class WANDBHook(Hook):
    """
    Wandb Hook
    """

    def __init__(self):
        super().__init__()
        self.log_key_list = ['train/sup_loss', 'train/unsup_loss', 'train/total_loss', 'train/util_ratio', 
                             'train/run_time', 'train/prefetch_time', 'lr',
                             'eval/top-1-acc', 'eval/precision', 'eval/recall', 'eval/F1']

    def before_run(self, algorithm):
        # job_id = '_'.join(algorithm.args.save_name.split('_')[:-1])
        name = algorithm.save_name
        project = algorithm.save_dir.split('/')[-1]

        # tags
        benchmark = f'benchmark: {project}'
        dataset = f'dataset: {algorithm.args.dataset}'
        data_setting = f'setting: {algorithm.args.dataset}_lb{algorithm.args.num_labels}_{algorithm.args.lb_imb_ratio}_ulb{algorithm.args.ulb_num_labels}_{algorithm.args.ulb_imb_ratio}'
        alg = f'alg: {algorithm.args.algorithm}'
        imb_alg = f'imb_alg: {algorithm.args.imb_algorithm}'
        tags = [benchmark, dataset, data_setting, alg, imb_alg] 
        if algorithm.args.resume:
            resume = 'auto'
        else:
            resume = 'never'
        # resume = 'never'
        
        save_dir = os.path.join(algorithm.args.save_dir, 'wandb', algorithm.args.save_name)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        self.run = wandb.init(name=name, 
                              tags=tags, 
                              config=algorithm.args.__dict__, 
                              project=project, 
                              resume=resume,
                              dir=save_dir)


    def after_train_step(self, algorithm):
        if self.every_n_iters(algorithm, algorithm.num_log_iter):
            log_dict = {}
            for key, item in algorithm.log_dict.items():
                if key in self.log_key_list:
                    log_dict[key] = item
            self.run.log(log_dict, step=algorithm.it)
    
        if self.every_n_iters(algorithm, algorithm.num_eval_iter):
            self.run.log({'eval/best-acc': algorithm.best_eval_acc}, step=algorithm.it)
    
    def after_run(self, algorithm):
        self.run.finish()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.


import torch 
import torch.nn as nn 
from torch.nn import functional as F

from .cross_entropy import ce_loss



def consistency_loss(logits, targets, name='ce', mask=None):
    """
    consistency regularization loss in semi-supervised learning.

    Args:
        logits: logit to calculate the loss on and back-propagation, usually being the strong-augmented unlabeled samples
        targets: pseudo-labels (either hard label or soft label)
        name: use cross-entropy ('ce') or mean-squared-error ('mse') to calculate loss
        mask: masks to mask-out samples when calculating the loss, usually being used as confidence-masking-out
    """

    assert name in ['ce', 'mse']
    # logits_w = logits_w.detach()
    if name == 'mse':
        probs = torch.softmax(logits, dim=-1)
        loss = F.mse_loss(probs, targets, reduction='none').mean(dim=1)
    else:
        loss = ce_loss(logits, targets, reduction='none')

    if mask is not None:
        # mask must not be boolean type
        loss = loss * mask

    return loss.mean()



class ConsistencyLoss(nn.Module):
    """
    Wrapper for consistency loss
    """
    def forward(self, logits, targets, name='ce', mask=None):
        return consistency_loss(logits, targets, name, mask)


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.


import torch 
import torch.nn as nn

from torch.nn import functional as F


def ce_loss(logits, targets, reduction='none'):
    """
    cross entropy loss in pytorch.

    Args:
        logits: logit values, shape=[Batch size, # of classes]
        targets: integer or vector, shape=[Batch size] or [Batch size, # of classes]
        # use_hard_labels: If True, targets have [Batch size] shape with int values. If False, the target is vector (default True)
        reduction: the reduction argument
    """
    if logits.shape == targets.shape:
        # one-hot target
        log_pred = F.log_softmax(logits, dim=-1)
        nll_loss = torch.sum(-targets * log_pred, dim=1)
        if reduction == 'none':
            return nll_loss
        else:
            return nll_loss.mean()
    else:
        log_pred = F.log_softmax(logits, dim=-1)
        return F.nll_loss(log_pred, targets, reduction=reduction)


class CELoss(nn.Module):
    """
    Wrapper for ce loss
    """
    def forward(self, logits, targets, reduction='none'):
        return ce_loss(logits, targets, reduction)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.


import torch
from semilearn.algorithms.hooks import MaskingHook

class AdaMatchThresholdingHook(MaskingHook):
    """
    Relative Confidence Thresholding in AdaMatch
    """

    @torch.no_grad()
    def masking(self, algorithm, logits_x_lb, logits_x_ulb, softmax_x_lb=True, softmax_x_ulb=True,  *args, **kwargs):
        if softmax_x_ulb:
            # probs_x_ulb = torch.softmax(logits_x_ulb.detach(), dim=-1)
            probs_x_ulb = algorithm.compute_prob(logits_x_ulb.detach())
        else:
            # logits is already probs
            probs_x_ulb = logits_x_ulb.detach()

        if softmax_x_lb:
            # probs_x_lb = torch.softmax(logits_x_lb.detach(), dim=-1)
            probs_x_lb = algorithm.compute_prob(logits_x_lb.detach())
        else:
            # logits is already probs
            probs_x_lb = logits_x_lb.detach()

        max_probs, _ = probs_x_lb.max(dim=-1)
        p_cutoff = max_probs.mean() * algorithm.p_cutoff
        max_probs, _ = probs_x_ulb.max(dim=-1)
        mask = max_probs.ge(p_cutoff).to(max_probs.dtype)
        return mask


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.


from .hook import Hook
from semilearn.core.utils import EMA


class EMAHook(Hook):
    """
    EMA model Hook for updating ema version of the model
    """

    def before_run(self, algorithm):
        algorithm.ema = EMA(algorithm.model, algorithm.ema_m)
        algorithm.ema.register()
        if algorithm.resume == True:
            algorithm.ema.load(algorithm.ema_model)

    def after_train_step(self, algorithm):
        if algorithm.ema is not None:
            algorithm.ema.update()
            algorithm.ema_model.load_state_dict(algorithm.model.state_dict())
            algorithm.ema_model.load_state_dict(algorithm.ema.shadow, strict=False)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch
import numpy as np
import torch.nn.functional as F
from semilearn.core import AlgorithmBase
from semilearn.core.utils import ALGORITHMS
from semilearn.algorithms.utils import SSL_Argument, str2bool, mixup_one_target


@ALGORITHMS.register('mixmatch')
class MixMatch(AlgorithmBase):
    """
        MixMatch algorithm (https://arxiv.org/abs/1905.02249).

        Args:
            - args (`argparse`):
                algorithm arguments
            - net_builder (`callable`):
                network loading function
            - tb_log (`TBLog`):
                tensorboard logger
            - logger (`logging.Logger`):
                logger to use
            - T (`float`):
                Temperature for pseudo-label sharpening
            - unsup_warm_up (`float`, *optional*, defaults to 0.4):
                Ramp up for weights for unsupervised loss
            - mixup_alpha (`float`, *optional*, defaults to 0.5):
                Hyper-parameter of mixup
            - mixup_manifold (`bool`, *optional*, defaults to `False`):
                Whether or not to use manifold mixup
    """
    def __init__(self, args, net_builder, tb_log=None, logger=None):
        super().__init__(args, net_builder, tb_log, logger)
        # mixmatch specified arguments
        self.init(T=args.T, unsup_warm_up=args.unsup_warm_up, mixup_alpha=args.mixup_alpha, mixup_manifold=args.mixup_manifold)

    def init(self, T, unsup_warm_up=0.01525, mixup_alpha=0.5, mixup_manifold=False):
        self.T = T
        self.unsup_warm_up = unsup_warm_up
        self.mixup_alpha = mixup_alpha
        self.mixup_manifold = mixup_manifold

    def train_step(self, x_lb, y_lb, x_ulb_w, x_ulb_s):
        num_lb = y_lb.shape[0]

        # inference and calculate sup/unsup losses
        with self.amp_cm():
            with torch.no_grad():
                self.bn_controller.freeze_bn(self.model)
                outs_x_ulb_w1 = self.model(x_ulb_w)
                logits_x_ulb_w1 = outs_x_ulb_w1['logits']
                feat_x_ulb_w1 = outs_x_ulb_w1['feat']

                # logits_x_ulb_w1 = self.model(x_ulb_w)
                outs_x_ulb_w2 = self.model(x_ulb_s)
                logits_x_ulb_w2 = outs_x_ulb_w2['logits']
                feat_x_ulb_w2 = outs_x_ulb_w2['feat']

                # logits_x_ulb_w2 = self.model(x_ulb_s)
                self.bn_controller.unfreeze_bn(self.model)
                
                # avg
                # avg_prob_x_ulb = (torch.softmax(logits_x_ulb_w1, dim=1) + torch.softmax(logits_x_ulb_w2, dim=1)) / 2
                avg_prob_x_ulb = (self.compute_prob(logits_x_ulb_w1) + self.compute_prob(logits_x_ulb_w2)) / 2
                # avg_prob_x_ulb = (avg_prob_x_ulb / avg_prob_x_ulb.sum(dim=-1, keepdim=True))
                # sharpening
                sharpen_prob_x_ulb = avg_prob_x_ulb ** (1 / self.T)
                sharpen_prob_x_ulb = (sharpen_prob_x_ulb / sharpen_prob_x_ulb.sum(dim=-1, keepdim=True)).detach()
            
            
            self.bn_controller.freeze_bn(self.model)
            outs_x_lb = self.model(x_lb)
            self.bn_controller.unfreeze_bn(self.model)
            feats_x_lb = outs_x_lb['feat']
            feat_dict = {'x_lb':feats_x_lb, 'x_ulb_w':feat_x_ulb_w1, 'x_ulb_s':feat_x_ulb_w2}

            # with torch.no_grad():
            # Pseudo Label
            input_labels = torch.cat([F.one_hot(y_lb, self.num_classes), sharpen_prob_x_ulb, sharpen_prob_x_ulb], dim=0)
            # Mix up
            if self.mixup_manifold:
                inputs = torch.cat((outs_x_lb['feat'], outs_x_ulb_w1['feat'], outs_x_ulb_w2['feat']))
            else:
                inputs = torch.cat([x_lb, x_ulb_w, x_ulb_s])
            mixed_x, mixed_y, _ = mixup_one_target(inputs, input_labels,
                                                   self.mixup_alpha,
                                                   is_bias=True)
            mixed_x = list(torch.split(mixed_x, num_lb))
            # mixed_x = interleave(mixed_x, num_lb)

            if self.mixup_manifold:
                logits = [self.model(mixed_x[0], only_fc=self.mixup_manifold)]
                # calculate BN for only the first batch
                self.bn_controller.freeze_bn(self.model)
                for ipt in mixed_x[1:]:
                    logits.append(self.model(ipt, only_fc=self.mixup_manifold))
                self.bn_controller.unfreeze_bn(self.model)
            else:
                logits = [self.model(mixed_x[0])['logits']]
                # calculate BN for only the first batch
                self.bn_controller.freeze_bn(self.model)
                for ipt in mixed_x[1:]:
                    logits.append(self.model(ipt)['logits'])
                self.bn_controller.unfreeze_bn(self.model)

            # put interleaved samples back
            # logits = interleave(logits, num_lb)

            logits_x = logits[0]
            logits_u = torch.cat(logits[1:], dim=0)

            sup_loss = self.ce_loss(logits_x, mixed_y[:num_lb], reduction='mean')
            unsup_loss = self.consistency_loss(logits_u, mixed_y[num_lb:], name='mse')

            # set ramp_up for lambda_u
            unsup_warmup = float(np.clip(self.it / (self.unsup_warm_up * self.num_train_iter), 0.0, 1.0))
            lambda_u = self.lambda_u * unsup_warmup

            total_loss = sup_loss + lambda_u * unsup_loss

        out_dict = self.process_out_dict(loss=total_loss, feat=feat_dict)
        log_dict = self.process_log_dict(sup_loss=sup_loss.item(), 
                                         unsup_loss=unsup_loss.item(), 
                                         total_loss=total_loss.item())
        return out_dict, log_dict


    @staticmethod
    def get_argument():
        return [
            SSL_Argument('--T', float, 0.5, 'parameter for Temperature Sharpening'),
            SSL_Argument('--unsup_warm_up', float, 1 / 64, 'ramp up ratio for unsupervised loss'),
            SSL_Argument('--mixup_alpha', float, 0.5, 'parameter for Beta distribution of Mix Up'),
            SSL_Argument('--mixup_manifold', str2bool, False, 'use manifold mixup (for nlp)'),
        ]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.


import torch

from semilearn.algorithms.hooks import MaskingHook

class DashThresholdingHook(MaskingHook):
    """
    Dynamic Threshold in DASH
    """
    
    def __init__(self, rho_min, gamma, C, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gamma = gamma
        self.C = C
        self.rho_update_cnt = 0
        self.rho_init = None
        self.rho_min = rho_min
        self.rho = None
    
    @torch.no_grad()
    def update(self, algorithm):
        if self.rho_init is None:
            self.rho_init = algorithm.rho_init

        # adjust rho every 10 epochs
        if algorithm.it % (10 * algorithm.num_iter_per_epoch) == 0:
            self.rho = self.C * (self.gamma ** -self.rho_update_cnt) * self.rho_init
            self.rho = max(self.rho, self.rho_min)
            self.rho_update_cnt += 1
        
        # use hard labels if rho reduced 0.05
        if self.rho == self.rho_min:
            algorithm.use_hard_label = True
        else:
            algorithm.use_hard_label = False 
    
    @torch.no_grad()
    def masking(self, algorithm, logits_x_ulb, *args, **kwargs):    
        self.update(algorithm)

        if algorithm.use_hard_label:
            pseudo_label = torch.argmax(logits_x_ulb, dim=-1).detach()
        else:
            # pseudo_label = torch.softmax(logits_x_ulb / algorithm.T, dim=-1).detach()
            pseudo_label = algorithm.compute_prob(logits_x_ulb.detach() / algorithm.T)
        loss_w = algorithm.ce_loss(logits_x_ulb, pseudo_label, reduction='none')
        mask = loss_w.le(self.rho).to(logits_x_ulb.dtype)
        return mask


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import torch.nn as nn 


class SSL_Argument(object):
    """
    Algrithm specific argument
    """
    def __init__(self, name, type, default, help=''):
        """
        Model specific arguments should be added via this class.
        """
        self.name = name
        self.type = type
        self.default = default
        self.help = help


def str2bool(v):
    """
    str to bool
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch 
import torch.nn as nn
import numpy as np
from inspect import signature

from .utils import TRASLogitsAdjCELoss, TRASKLLoss
from semilearn.core import ImbAlgorithmBase
from semilearn.algorithms.utils import SSL_Argument
from semilearn.core.utils import IMB_ALGORITHMS


class TRASNet(nn.Module):
    """
        Transfer & Share algorithm (https://arxiv.org/abs/2205.13358).

        Args:
            - args (`argparse`):
                algorithm arguments
            - net_builder (`callable`):
                network loading function
            - tb_log (`TBLog`):
                tensorboard logger
            - logger (`logging.Logger`):
                logger to use
            - tras_A
                A parameter in TRAS
            - tras_B
                B parameter  in TRAS
            - tras_tro:
                tro parameter in TRAS
            - tras_warmup_epochs:
                TRAS warmup epochs
    """
    def __init__(self, backbone, num_classes):
        super().__init__()
        self.backbone = backbone
        self.num_features = backbone.num_features

        # auxilary classifier
        self.aux_classifier = nn.Linear(self.backbone.num_features, num_classes)
    
    def forward(self, x, **kwargs):
        results_dict = self.backbone(x, **kwargs)
        results_dict['logits_aux'] = self.aux_classifier(results_dict['feat'])
        return results_dict

    def group_matcher(self, coarse=False):
        if hasattr(self.backbone, 'backbone'):
            # TODO: better way
            matcher = self.backbone.backbone.group_matcher(coarse, prefix='backbone.backbone')
        else:
            matcher = self.backbone.group_matcher(coarse, prefix='backbone.')
        return matcher


@IMB_ALGORITHMS.register('tras')
class TRAS(ImbAlgorithmBase):
    def __init__(self, args, **kwargs):
        self.imb_init(A=args.tras_A, B=args.tras_B, tro=args.tras_tro, warmup_epochs=args.tras_warmup_epochs)
        super().__init__(args, **kwargs)
        assert args.algorithm == 'fixmatch', "Adsh only supports FixMatch as the base algorithm."

        # comput lb imb ratio
        lb_class_dist = [0 for _ in range(self.num_classes)]
        for c in  self.dataset_dict['train_lb'].targets:
            lb_class_dist[c] += 1
        lb_class_dist = np.array(lb_class_dist)
        self.lb_class_dist = torch.from_numpy(np.min(lb_class_dist) / lb_class_dist)
        
        # TODO: better ways
        self.model = TRASNet(self.model, num_classes=self.num_classes)
        self.ema_model = TRASNet(self.ema_model, num_classes=self.num_classes)
        self.ema_model.load_state_dict(self.model.state_dict())
        self.optimizer, self.scheduler = self.set_optimizer()

        # compute T logits
        self.la = torch.log(self.lb_class_dist ** self.tro + 1e-12).to(self.gpu)
        T_logit = torch.softmax(-self.la / 1, dim=0)
        self.T_logit = self.A * T_logit + self.B

        # crete tras ce loss
        self.tras_ce_loss = TRASLogitsAdjCELoss(la=self.la)
        self.tras_kl_loss = TRASKLLoss()

    def imb_init(self, A, B, tro, warmup_epochs):
        self.A = A
        self.B = B
        self.tro = tro
        self.warmup_epochs = warmup_epochs

    def process_batch(self, **kwargs):
        # get core algorithm parameters
        input_args = signature(super().train_step).parameters
        input_args = list(input_args.keys())
        return super().process_batch(input_args=input_args, **kwargs)

    def train_step(self, *args, **kwargs):
        out_dict, log_dict = super().train_step(*args, **kwargs)

        if self.epoch < self.warmup_epochs:
            return out_dict, log_dict
        
        # get features
        feats_x_lb = out_dict['feat']['x_lb']
        feats_x_ulb_w = out_dict['feat']['x_ulb_w']
        feats_x_ulb_s = out_dict['feat']['x_ulb_s']
        if isinstance(feats_x_ulb_s, list):
            feats_x_ulb_s = feats_x_ulb_s[0]
        
        # get logits
        logits_x_lb = self.model.module.aux_classifier(feats_x_lb)
        logits_x_ulb_s = self.model.module.aux_classifier(feats_x_ulb_s)
        with torch.no_grad():
            logits_x_ulb_w = self.model.module.aux_classifier(feats_x_ulb_w)
        
        # compute supervised loss 
        tras_sup_loss = self.tras_ce_loss(logits_x_lb, kwargs['y_lb'], reduction='mean')

        # compute mask
        probs_x_ulb_w = torch.softmax(logits_x_ulb_w.detach(), dim=-1)
        mask = self.call_hook("masking", "MaskingHook", logits_x_ulb=probs_x_ulb_w, softmax_x_ulb=False)

        # generate unlabeled targets using pseudo label hook
        pseudo_label = self.call_hook("gen_ulb_targets", "PseudoLabelingHook", 
                                        logits=probs_x_ulb_w,
                                        use_hard_label=self.use_hard_label,
                                        T=self.T,
                                        softmax=False)

        la_u = self.la.expand([logits_x_ulb_s.size(0), self.num_classes])
        la_u = (la_u.t() * self.T_logit[pseudo_label].cuda()).t()

        # TRAS loss of unlabeled
        tras_unsup_loss = self.tras_kl_loss(logits_x_ulb_s, logits_x_ulb_w.detach()-la_u, 1, mask)

        tras_loss = tras_sup_loss + tras_unsup_loss

        out_dict['loss'] += tras_loss
        log_dict['train/tras_loss'] = tras_loss.item()
        return out_dict, log_dict

    def compute_prob(self, logits):
        return super().compute_prob(logits - self.la)

    def evaluate(self, eval_dest='eval', out_key='logits', return_logits=False):
        if self.epoch < self.warmup_epochs:
            out_key = 'logits'
        else:
            out_key = 'logits_aux'
        return super().evaluate(eval_dest=eval_dest, out_key=out_key, return_logits=return_logits)

    @staticmethod
    def get_argument():
        return [
            SSL_Argument('--tras_warmup_epochs', int, 10),
            SSL_Argument('--tras_A', int, 2),
            SSL_Argument('--tras_B', int, 2),
            SSL_Argument('--tras_tro', float, 1.0),
        ]

        


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.


from cProfile import label
from collections import defaultdict
import torch

from semilearn.algorithms.hooks import PseudoLabelingHook


class DASOFeatureQueue:
    def __init__(self, num_classes, feat_dim, queue_length, classwise_max_size=None, bal_queue=True):
        self.num_classes = num_classes
        self.feat_dim = feat_dim
        self.max_size = queue_length

        self.bank = defaultdict(lambda: torch.empty(0, self.feat_dim))
        self.prototypes = torch.zeros(self.num_classes, self.feat_dim)
        self.classwise_max_size = classwise_max_size
        self.bal_queue = bal_queue

    def enqueue(self, features: torch.Tensor, labels: torch.Tensor): 
        if not self.prototypes.is_cuda:
            self.prototypes = self.prototypes.to(features.device)

        for idx in range(self.num_classes):
            
            # per class max size
            max_size = (self.classwise_max_size[idx] * 5) if self.classwise_max_size is not None else self.max_size
            if self.bal_queue:
                max_size = self.max_size

            # select features by label
            cls_inds = torch.where(labels == idx)[0]
            if len(cls_inds):
                with torch.no_grad():
                    # push to the memory bank
                    feats_selected = features[cls_inds]
                    self.bank[idx] = torch.cat([self.bank[idx], feats_selected.cpu()], 0)

                    # fixed size
                    current_size = len(self.bank[idx])
                    if current_size > max_size:
                        self.bank[idx] = self.bank[idx][current_size - max_size:]

                    # update prototypes
                    self.prototypes[idx, :] = self.bank[idx].mean(0).to(features.device)
    

class DASOPseudoLabelingHook(PseudoLabelingHook):
    def __init__(self, num_classes, T_dist, with_dist_aware, interp_alpha):
        super().__init__()
        self.num_classes = num_classes
        self.T_dist = T_dist
        self.with_dist_aware = with_dist_aware
        self.interp_alpha = interp_alpha

        self.pseudo_label_list = []
        self.pseudo_label_dist = [0 for i in range(self.num_classes)] 

    def push_pl_list(self, pl_list):
        self.pseudo_label_list.append(pl_list)

    def update_pl_dist(self):
        if len(self.pseudo_label_list) == 0: return

        pl_total_list = torch.cat(self.pseudo_label_list, 0)
        for class_ind in range(self.num_classes):
            pl_row_inds = torch.where(pl_total_list == class_ind)[0]
            self.pseudo_label_dist[class_ind] = len(pl_row_inds)
        self.pseudo_label_list = []

    def get_pl_dist(self, normalize=True):
        if isinstance(self.pseudo_label_dist, list):
            pl_dist = torch.Tensor(self.pseudo_label_dist).float()
        else:
            pl_dist = self.pseudo_label_dist.float()
        if normalize:
            pl_dist = pl_dist / pl_dist.sum()
        return pl_dist


    @torch.no_grad()
    def gen_ulb_targets(self, 
                        algorithm, 
                        logits, 
                        use_hard_label=True, 
                        T=1.0,
                        softmax=True, # whether to compute softmax for logits, input must be logits
                        label_smoothing=0.0):

        if algorithm.it  < algorithm.num_pretrain_iter:
            pseudo_label = torch.argmax(logits.detach(), dim=-1)
            # update pseudo label list
            self.push_pl_list(pseudo_label.detach().cpu())
            return super().gen_ulb_targets(algorithm=algorithm, logits=logits, use_hard_label=use_hard_label, T=T, softmax=softmax, label_smoothing=label_smoothing)

        logits = logits.detach()
        if softmax:
            # probs = torch.softmax(logits, dim=1)
            probs = self.compute_prob(logits)
        else:
            probs = logits
        pseudo_label = torch.argmax(probs.detach(), dim=-1)

        # compute the mix
        current_pl_dist = self.get_pl_dist().to(probs.device)  # (1, C)
        current_pl_dist = current_pl_dist**(1. / algorithm.T_dist)
        current_pl_dist = current_pl_dist / current_pl_dist.sum()
        current_pl_dist = current_pl_dist / current_pl_dist.max()  # MIXUP
        
        pred_to_dist = current_pl_dist[pseudo_label].view(-1, 1)  # (B, )

        if not algorithm.with_dist_aware:
            pred_to_dist = algorithm.interp_alpha  # override to fixed constant

        # pl mixup
        probs_mixup = (1. - pred_to_dist) * probs + pred_to_dist * algorithm.probs_sim

        pseudo_label = torch.argmax(probs_mixup.detach(), dim=-1) 

        # update pseudo label list
        self.push_pl_list(pseudo_label.detach().cpu())

        if use_hard_label:
            return pseudo_label
        else:
            return probs_mixup
    
    def after_train_step(self, algorithm):
        if self.every_n_iters(algorithm, algorithm.num_pl_dist_iter):
            self.update_pl_dist()


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import importlib

__all__ = [
    'ALGORITHMS',
    'IMB_ALGORITHMS',
]

class Register:
    def __init__(self, registry_name):
        self._dict = {}
        self._name = registry_name

    def __setitem__(self, key, value):
        if not callable(value):
            raise Exception(f"Value of a Registry must be a callable!\nValue: {value}")
        if key is None:
            key = value.__name__
        if key in self._dict:
            print("Key %s already in registry %s." % (key, self._name))
        self._dict[key] = value

    def register(self, target):
        """Decorator to register a function or class."""

        def add(key, value):
            self[key] = value
            return value

        if callable(target):
            # @reg.register
            return add(None, target)
        # @reg.register('alias')
        return lambda x: add(target, x)

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def keys(self):
        """key"""
        return self._dict.keys()


ALGORITHMS = Register('algorithms')
IMB_ALGORITHMS = Register('imb_algorithms')



def _handle_errors(errors):
    """
    Log out and possibly reraise errors during import.
    """
    if not errors:
        return
    
    for name, err in errors:
        print("Module {} import failed: {}".format(name, err))


ALL_MODULES = [
    # NOTE: add all algorithms here
    ('semilearn.algorithms', ['adamatch', 'comatch', 'crmatch', 'dash', 'fixmatch', 'flexmatch', 'fullysupervised', 'meanteacher',
                              'mixmatch', 'pimodel', 'pseudolabel', 'remixmatch', 'simmatch', 'uda', 'vat', 'softmatch', 'freematch', 'defixmatch']),
    ('semilearn.imb_algorithms', ['abc', 'cossl', 'adsh', 'crest', 'darp', 'daso', 'debiaspl', 'saw', 'tras'])
]


def import_all_modules_for_register():
    """
    Import all modules for register.
    """
    all_modules = ALL_MODULES
    errors = []
    for base_dir, modules in all_modules:
        for name in modules:
            try:
                if base_dir != "":
                    full_name = base_dir + "." + name
                else:
                    full_name = name
                importlib.import_module(full_name)
            except ImportError as error:
                errors.append((name, error))
    _handle_errors(errors)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# Ref: https://github.com/bbuing9/DARP/blob/64a94b46b5fcf3307178c2cb2f97fe051c43e500/common.py#L107

import math
import torch
import numpy as np
from scipy import optimize
from semilearn.algorithms.hooks import PseudoLabelingHook
from semilearn.algorithms.utils import smooth_targets


class DARPPseudoLabelingHook(PseudoLabelingHook):
    def __init__(self, warmup_epochs, alpha, iter_T, num_refine_iter, dataset_len, num_classes, target_disb):
        super(DARPPseudoLabelingHook, self).__init__()
        self.warmup_epochs = warmup_epochs
        self.alpha = alpha
        self.iter_T = iter_T
        self.num_refine_iter = num_refine_iter
        self.num_classes = num_classes
        self.dataset_len = dataset_len
        self.target_disb = torch.tensor(target_disb)

        self.pseudo_orig = torch.ones(self.dataset_len, self.num_classes) / self.num_classes
        self.pseudo_refine = torch.ones(self.dataset_len, self.num_classes) / self.num_classes

    @torch.no_grad()
    def refine_pseudo_labels(self, idx_u, targets_u, batch_idx, epoch):
        # Update the saved predictions with current one
        self.pseudo_orig[idx_u, :] = targets_u.data.cpu().to(self.pseudo_orig.dtype)
        self.pseudo_orig_backup = self.pseudo_orig.clone()

        if epoch > self.warmup_epochs:
            if batch_idx % self.num_refine_iter == 0:
                # Iterative normalization
                targets_u, weights_u = self.estimate_pseudo(self.target_disb, self.pseudo_orig)
                scale_term = targets_u * weights_u.reshape(1, -1)
                self.pseudo_orig = (self.pseudo_orig * scale_term + 1e-6) \
                                   / (self.pseudo_orig * scale_term + 1e-6).sum(dim=1, keepdim=True)

                opt_res = self.opt_solver(self.pseudo_orig, self.target_disb, self.iter_T)

                # Updated pseudo-labels are saved
                self.pseudo_refine = opt_res

                # Select
                targets_u = opt_res[idx_u.to(opt_res.device)].detach().cuda()
                self.pseudo_orig = self.pseudo_orig_backup
            else:
                # Using previously saved pseudo-labels
                targets_u = self.pseudo_refine[idx_u.to(self.pseudo_refine.device)].cuda()

        return targets_u

    @torch.no_grad()
    def estimate_pseudo(self, q_y, saved_q):
        pseudo_labels = torch.zeros(len(saved_q), self.num_classes)
        k_probs = torch.zeros(self.num_classes)

        for i in range(1, self.num_classes + 1):
            i = self.num_classes - i
            num_i = int(self.alpha * q_y[i])
            sorted_probs, idx = saved_q[:, i].sort(dim=0, descending=True)
            pseudo_labels[idx[: num_i], i] = 1
            k_probs[i] = sorted_probs[:num_i].sum()

        return pseudo_labels, (q_y + 1e-6) / (k_probs + 1e-6)

    @torch.no_grad()
    def f(self, x, a, b, c, d):
        return np.sum(a * b * np.exp(-1 * x / c)) - d

    @torch.no_grad()
    def opt_solver(self, probs, target_distb, num_iter, num_newton=30):
        entropy = (-1 * probs * torch.log(probs + 1e-6)).sum(1)
        weights = (1 / entropy)
        N, K = probs.size(0), probs.size(1)

        A, w, lam, nu, r, c = probs.numpy(), weights.numpy(), np.ones(N), np.ones(K), np.ones(N), target_distb.numpy()
        A_e = A / math.e
        X = np.exp(-1 * lam / w)
        Y = np.exp(-1 * nu.reshape(1, -1) / w.reshape(-1, 1))
        prev_Y = np.zeros(K)
        X_t, Y_t = X, Y

        for n in range(num_iter):
            # Normalization
            denom = np.sum(A_e * Y_t, 1)
            X_t = r / denom

            # Newton method
            Y_t = np.zeros(K)
            for i in range(K):
                Y_t[i] = optimize.newton(self.f, prev_Y[i], maxiter=num_newton, args=(A_e[:, i], X_t, w, c[i]), tol=1.0e-01)
            prev_Y = Y_t
            Y_t = np.exp(-1 * Y_t.reshape(1, -1) / w.reshape(-1, 1))

        denom = np.sum(A_e * Y_t, 1)
        X_t = r / denom
        M = torch.Tensor(A_e * X_t.reshape(-1, 1) * Y_t)

        return M

    @torch.no_grad()
    def gen_ulb_targets(self, algorithm, logits, use_hard_label=True, T=1.0,
                        softmax=True,  # whether to compute softmax for logits, input must be logits
                        label_smoothing=0.0):
        logits = logits.detach()
        if softmax:
            # probs = torch.softmax(logits / T, dim=1)
            probs = algorithm.compute_prob(logits / T)
        else:
            probs = logits

        targets_u = self.refine_pseudo_labels(algorithm.idx_ulb, probs,
                                              algorithm.it, algorithm.epoch)

        if use_hard_label:
            # return hard label directly
            pseudo_label = torch.argmax(targets_u, dim=-1)

            if label_smoothing:
                pseudo_label = smooth_targets(logits, pseudo_label, label_smoothing)
            return pseudo_label
        else:  # return soft label
            return targets_u


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
import copy
import torch
import numpy as np

from .utils import ProgressiveDistAlignEMAHook, CReSTCheckpointHook, CReSTLoggingHook

from semilearn.core import ImbAlgorithmBase
from semilearn.core.utils import get_dataset, get_data_loader, send_model_cuda, IMB_ALGORITHMS
from semilearn.algorithms.utils import SSL_Argument, str2bool


@IMB_ALGORITHMS.register('crest')
class CReST(ImbAlgorithmBase):
    """
        CReST algorithm (https://arxiv.org/abs/2102.09559).

        Args:
            - args (`argparse`):
                algorithm arguments
            - net_builder (`callable`):
                network loading function
            - tb_log (`TBLog`):
                tensorboard logger
            - logger (`logging.Logger`):
                logger to use
            - crest_num_gens (int):
                number of generations for crest
            - crest_dist_align_t (float):
                t parameter in dist align
            - crest_pro_dist_align (bool):
                flag of using progressive dist align
            - crest_alpha (float):
                alpha parameter for crest
    """
    def __init__(self, args, net_builder, tb_log=None, logger=None, **kwargs):
        self.imb_init(num_gens=args.crest_num_gens, dist_align_t=args.crest_dist_align_t, pro_dist_align=args.crest_pro_dist_align, sampling_alpha=args.crest_alpha)
        super(CReST, self).__init__(args, net_builder, tb_log, logger, **kwargs)

    def imb_init(self, num_gens=6, dist_align_t=0.5, pro_dist_align=True, sampling_alpha=3):
        self.num_gens = num_gens
        self.dist_align_t = dist_align_t
        self.pro_dist_align = pro_dist_align
        self.sampling_alpha = sampling_alpha
        self.start_gen = 0
        self.pseudo_label_list = None
        self.best_gen = 0 
        self.best_gen_eval_acc = 0.0
    
    def set_hooks(self):
        super().set_hooks()
        
        # reset checkpoint hook
        self.register_hook(CReSTCheckpointHook(), "CheckpointHook", "HIGH")
        self.register_hook(CReSTLoggingHook(), "LoggingHook", "LOW")

        lb_class_dist = [0 for _ in range(self.num_classes)]
        for c in  self.dataset_dict['train_lb'].targets:
            lb_class_dist[c] += 1
        lb_class_dist = np.array(lb_class_dist)
        lb_class_dist = lb_class_dist / lb_class_dist.sum()
        self.lb_class_dist = lb_class_dist

        # get ground truth distribution
        if self.pro_dist_align:
            self.register_hook(
                ProgressiveDistAlignEMAHook(num_classes=self.num_classes, p_target_type='gt', p_target=lb_class_dist), 
                "DistAlignHook")

    def get_split(self, lb_data, lb_targets, eval_ulb_data, eval_ulb_targets, pseudo_label_list=None):
        if pseudo_label_list is not None and len(pseudo_label_list):
            data_picked = []
            targets_picked = []

            lb_class_dist = self.lb_class_dist
            sorted_class = np.argsort(lb_class_dist)[::-1]
            class_imb_ratio = lb_class_dist[sorted_class][0] / lb_class_dist[sorted_class[-1]]  # self.lb_imb_ratio
            class_imb_ratio = 1. / class_imb_ratio
            mu = np.math.pow(class_imb_ratio, 1 / (self.num_classes - 1))

            for c in sorted_class:
                num_picked = int(
                    len(pseudo_label_list[c]) *
                    np.math.pow(np.math.pow(mu, (self.num_classes - 1) - c), 1 / self.sampling_alpha))  # this is correct!!!
                idx_picked = pseudo_label_list[c][:num_picked]

                try:
                    if len(idx_picked) > 0:
                        data_picked.append(eval_ulb_data[idx_picked])
                        targets_picked.append(np.ones_like(eval_ulb_targets[idx_picked]) * c)
                        print('class {} is added {} pseudo labels'.format(c, num_picked))
                except:
                    continue
            data_picked.append(lb_data)
            targets_picked.append(lb_targets)
            lb_data = np.concatenate(data_picked, axis=0)
            lb_targets = np.concatenate(targets_picked, axis=0)
        else:
            self.print_fn('Labeled data not update')
        return lb_data, lb_targets

    def set_dataset(self, pseudo_label_list=None):
        if self.rank != 0 and self.distributed:
            torch.distributed.barrier()
        
        # set include_lb_to_ulb to False
        dataset_dict = get_dataset(self.args, self.algorithm, self.args.dataset, self.args.num_labels, self.args.num_classes, self.args.data_dir, include_lb_to_ulb=False)
        # eval_ulb
        dataset_dict['eval_ulb'] = copy.deepcopy(dataset_dict['train_ulb'])
        dataset_dict['eval_ulb'].is_ulb = False

        # add pseudo labels into lb
        lb_data, lb_targets = dataset_dict['train_lb'].data, dataset_dict['train_lb'].targets
        eval_ulb_data, eval_ulb_targets = dataset_dict['eval_ulb'].data, dataset_dict['eval_ulb'].targets
        lb_data, lb_targets = self.get_split(lb_data, lb_targets, eval_ulb_data, eval_ulb_targets, pseudo_label_list)
        dataset_dict['train_lb'].data = lb_data
        dataset_dict['train_lb'].targets = lb_targets

        self.args.ulb_dest_len = len(dataset_dict['train_ulb']) if dataset_dict['train_ulb'] is not None else 0
        self.args.lb_dest_len = len(dataset_dict['train_lb'])
        self.print_fn("unlabeled data number: {}, labeled data number {}, unlabeled eval data number {}".format(self.args.ulb_dest_len, self.args.lb_dest_len, len(dataset_dict['eval_ulb'])))
        
        
        if self.rank == 0 and self.distributed:
            torch.distributed.barrier()
        return dataset_dict
    
    def set_data_loader(self):
        loader_dict = super().set_data_loader()
        
        # add unlabeled evaluation data loader
        loader_dict['eval_ulb'] = get_data_loader(self.args,
                                                  self.dataset_dict['eval_ulb'],
                                                  self.args.eval_batch_size,
                                                  data_sampler=None,
                                                  shuffle=False,
                                                  num_workers=self.args.num_workers,
                                                  drop_last=False)

        return loader_dict

    def re_init(self):
        self.it = 0
        self.best_gen_eval_acc = 0.0
        self.ema = None
        
        # build dataset with pseudo label list
        self.dataset_dict = self.set_dataset(self.pseudo_label_list)

        # build model and ema_model
        self.model = self.set_model()
        self.ema_model = self.set_ema_model()
        self.model = send_model_cuda(self.args, self.model)
        self.ema_model = send_model_cuda(self.args, self.ema_model)

        # build optimizer and scheduler
        self.optimizer, self.scheduler = self.set_optimizer()

        # build data loader
        self.loader_dict = self.set_data_loader()

    def train(self):

        # EMA Init
        self.model.train()

        for gen in range(self.start_gen, self.num_gens):
            self.gen = gen

            # before train generation
            if self.pro_dist_align:
                cur = self.gen / ( self.num_gens - 1)
                self.cur_dist_align_t = (1.0 - cur) * 1.0 + cur * self.dist_align_t
            else:
                self.cur_dist_align_t = self.dist_align_t

            # reinit every generation
            if self.gen > 0:
                self.re_init()
            
            self.call_hook("before_run")


            for epoch in range(self.start_epoch, self.epochs):
                self.epoch = epoch
                
                # prevent the training iterations exceed args.num_train_iter
                if self.it >= self.num_train_iter:
                    break

                self.call_hook("before_train_epoch")
                for data_lb, data_ulb in zip(self.loader_dict['train_lb'],
                                             self.loader_dict['train_ulb']):
                    # prevent the training iterations exceed args.num_train_iter
                    if self.it >= self.num_train_iter:
                        break

                    self.call_hook("before_train_step")
                    # NOTE: progressive dist align will be called inside each train_step in core algorithms
                    self.out_dict, self.log_dict = self.train_step(**self.process_batch(**data_lb, **data_ulb))
                    self.call_hook("after_train_step")
                    self.it += 1

                self.call_hook("after_train_epoch")
            
            # after train generation
            eval_dict = {'eval/best_acc': self.best_gen_eval_acc, 'eval/best_it': self.best_it}
            for key, item in eval_dict.items():
                self.print_fn(f"CReST Generation {gen}, Model result - {key} : {item}")

            self.print_fn(f"Generation {self.gen} finished, updating pseudo label list")
            ulb_logits = self.evaluate('eval_ulb', return_logits=True)['eval_ulb/logits']
            if isinstance(ulb_logits, np.ndarray):
                ulb_logits = torch.from_numpy(ulb_logits)
            ulb_score, ulb_pred = torch.max(torch.softmax(ulb_logits, dim=1), dim=1)
            self.pseudo_label_list = []
            for c in range(self.num_classes):
                idx_gather = torch.where(ulb_pred == c)[0]
                if len(idx_gather) == 0:
                    self.pseudo_label_list.append([])
                    continue
                score_gather = ulb_score[idx_gather]
                score_sorted_idx = torch.argsort(score_gather, descending=True)
                idx_gather = idx_gather[score_sorted_idx]
                self.pseudo_label_list.append(idx_gather.numpy())

            self.it = 0 
        
        self.call_hook("after_run")

    def get_save_dict(self):
        save_dict = super().get_save_dict()
        # additional saving arguments
        save_dict['gen'] = self.gen
        if self.pro_dist_align:
            save_dict['p_model'] = self.hooks_dict['DistAlignHook'].p_model.cpu()
            save_dict['p_target'] = self.hooks_dict['DistAlignHook'].p_target.cpu()
        return save_dict

    def load_model(self, load_path):
        checkpoint = super().load_model(load_path)
        self.gen = checkpoint['gen']
        self.start_gen = checkpoint['gen']
        if self.pro_dist_align:
            self.hooks_dict['DistAlignHook'].p_model = checkpoint['p_model'].cuda(self.args.gpu)
            self.hooks_dict['DistAlignHook'].p_target = checkpoint['p_target'].cuda(self.args.gpu)
        return checkpoint

    @staticmethod
    def get_argument():
        return [
            SSL_Argument('--crest_num_gens', int, 6),
            SSL_Argument('--crest_dist_align_t', float, 0.5),
            SSL_Argument('--crest_pro_dist_align', str2bool, True),
            SSL_Argument('--crest_alpha', float, 3),
        ]

IMB_ALGORITHMS['crest+'] = CReST



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .algorithmbase import AlgorithmBase, ImbAlgorithmBase
from .utils.registry import import_all_modules_for_register

import_all_modules_for_register()


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch
from .hook import Hook


class ParamUpdateHook(Hook):
    """
    Parameter Update Hook

    necessary for update the model parameters
    """
    
    def before_train_step(self, algorithm):
        if hasattr(algorithm, 'start_run'):
            torch.cuda.synchronize()
            algorithm.start_run.record()

    # call after each train_step to update parameters
    def after_train_step(self, algorithm):
        loss = algorithm.out_dict['loss']
        # algorithm.optimizer.zero_grad()
        # update parameters
        if algorithm.use_amp:
            algorithm.loss_scaler.scale(loss).backward()
            if (algorithm.clip_grad > 0):
                algorithm.loss_scaler.unscale_(algorithm.optimizer)
                torch.nn.utils.clip_grad_norm_(algorithm.model.parameters(), algorithm.clip_grad)
            algorithm.loss_scaler.step(algorithm.optimizer)
            algorithm.loss_scaler.update()
        else:
            loss.backward()
            if (algorithm.clip_grad > 0):
                torch.nn.utils.clip_grad_norm_(algorithm.model.parameters(), algorithm.clip_grad)
            algorithm.optimizer.step()

        if algorithm.scheduler is not None:
            algorithm.scheduler.step()
        algorithm.model.zero_grad()

        if hasattr(algorithm, 'end_run'):
            algorithm.end_run.record()
            torch.cuda.synchronize()
            algorithm.log_dict['train/run_time'] = algorithm.start_run.elapsed_time(algorithm.end_run) / 1000.




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch

from semilearn.core.hooks import Hook
from semilearn.core.criterions import ConsistencyLoss, consistency_loss

class DebiasPLHook(Hook):
    def before_train_step(self, algorithm):
        algorithm.consistency_loss.set_param(algorithm.p_hat)


class DebiasPLConsistencyLoss(ConsistencyLoss):
    def __init__(self, tau=0.4):
        super().__init__()
        self.tau = 0.4

    def set_param(self, p):
        self.p_hat = p

    def forward(self, logits, targets, name='ce', mask=None):
        return consistency_loss(logits + self.tau * torch.log(self.p_hat), targets, name, mask)


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from semilearn.core import AlgorithmBase
from semilearn.core.utils import ALGORITHMS
from semilearn.algorithms.hooks import DistAlignEMAHook 
from semilearn.algorithms.utils import str2bool, mixup_one_target, SSL_Argument


class ReMixMatch_Net(nn.Module):
    def __init__(self, base, use_rot=True):
        super(ReMixMatch_Net, self).__init__()
        self.backbone = base
        self.num_features = base.num_features

        if use_rot:
            self.rot_classifier = nn.Linear(self.num_features, 4)

    def forward(self, x, use_rot=False, **kwargs):
        if not use_rot:
            return self.backbone(x, **kwargs)
        
        feat = self.backbone(x, only_feat=True)
        logits = self.backbone(feat, only_fc=True)
        logits_rot = self.rot_classifier(feat)
        return {'logits':logits, 'logits_rot':logits_rot, 'feat':feat}
    
    def init(self, m):
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='leaky_relu')
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1)
            m.bias.data.zero_()
        elif isinstance(m, nn.Linear):
            nn.init.xavier_normal_(m.weight.data)
            m.bias.data.zero_()
    
    def group_matcher(self, coarse=False):
        matcher = self.backbone.group_matcher(coarse, prefix='backbone.')
        return matcher


@ALGORITHMS.register('remixmatch')
class ReMixMatch(AlgorithmBase):
    """
    ReMixMatch algorithm (https://arxiv.org/abs/1911.09785).

    Args:
        - args (`argparse`):
            algorithm arguments
        - net_builder (`callable`):
            network loading function
        - tb_log (`TBLog`):
            tensorboard logger
        - logger (`logging.Logger`):
            logger to use
        - T (`float`):
            Temperature for pseudo-label sharpening
        - unsup_warm_up (`float`, *optional*, defaults to 0.4):
            Ramp up for weights for unsupervised loss
        - mixup_alpha (`float`, *optional*, defaults to 0.5):
            Hyper-parameter of mixup
        - mixup_manifold (`bool`, *optional*, defaults to `False`):
            Whether or not to use manifold mixup
        - rot_loss_ratio ('float',  *optional*, defaults to 0.5):
            rotation loss weight
        - kl_loss_ratio ('float',  *optional*, defaults to 0.5):
            kl loss weight
    """
    def __init__(self, args, net_builder, tb_log=None, logger=None):
        # remixmatch specified arguments
        self.lambda_rot = args.rot_loss_ratio
        self.lambda_kl = args.kl_loss_ratio
        self.use_rot = self.lambda_rot > 0
        super().__init__(args, net_builder,  tb_log, logger)
        self.init(T=args.T, unsup_warm_up=args.unsup_warm_up, mixup_alpha=args.mixup_alpha, mixup_manifold=args.mixup_manifold)

    def init(self, T, unsup_warm_up=0.4, mixup_alpha=0.5, mixup_manifold=False):
        self.T = T
        self.unsup_warm_up = unsup_warm_up
        self.mixup_alpha = mixup_alpha
        self.mixup_manifold = mixup_manifold

    def set_hooks(self):
        lb_class_dist = [0 for _ in range(self.num_classes)]
        for c in  self.dataset_dict['train_lb'].targets:
            lb_class_dist[c] += 1
        lb_class_dist = np.array(lb_class_dist)
        lb_class_dist = lb_class_dist / lb_class_dist.sum()
        self.register_hook(
            DistAlignEMAHook(num_classes=self.num_classes, p_target_type='gt', p_target=lb_class_dist), 
            "DistAlignHook")
        super().set_hooks()

    def set_model(self):
        model = super().set_model()
        model = ReMixMatch_Net(model, self.use_rot)
        return model
    
    def set_ema_model(self):
        """
        initialize ema model from model
        """
        ema_model = self.net_builder(num_classes=self.num_classes)
        ema_model = ReMixMatch_Net(ema_model, self.use_rot)
        ema_model.load_state_dict(self.check_prefix_state_dict(self.model.state_dict()))
        return ema_model

    def train_step(self, x_lb, y_lb, x_ulb_w, x_ulb_s_0, x_ulb_s_1, x_ulb_s_0_rot=None, rot_v=None):
        num_lb = y_lb.shape[0]

        # inference and calculate sup/unsup losses
        with self.amp_cm():
            with torch.no_grad():
                self.bn_controller.freeze_bn(self.model)
                # logits_x_lb = self.model(x_lb)[0]

                outs_x_ulb_w = self.model(x_ulb_w)
                logits_x_ulb_w = outs_x_ulb_w['logits']
                feats_x_ulb_w = outs_x_ulb_w['feat']
                # logits_x_ulb_w = self.model(x_ulb_w)
                # logits_x_ulb_s1 = self.model(x_ulb_s1)[0]
                # logits_x_ulb_s2 = self.model(x_ulb_s2)[0]
                self.bn_controller.unfreeze_bn(self.model)

                prob_x_ulb = self.call_hook("dist_align", "DistAlignHook", probs_x_ulb=self.compute_prob(logits_x_ulb_w))
                sharpen_prob_x_ulb = prob_x_ulb ** (1 / self.T)
                sharpen_prob_x_ulb = (sharpen_prob_x_ulb / sharpen_prob_x_ulb.sum(dim=-1, keepdim=True)).detach()

            
            self.bn_controller.freeze_bn(self.model)
            outs_x_lb = self.model(x_lb)
            outs_x_ulb_s_0 = self.model(x_ulb_s_0)
            outs_x_ulb_s_1 = self.model(x_ulb_s_1)
            self.bn_controller.unfreeze_bn(self.model)

            feat_dict = {'x_lb':outs_x_lb['feat'], 'x_ulb_w':feats_x_ulb_w, 'x_ulb_s':[outs_x_ulb_s_0['feat'], outs_x_ulb_s_1['feat']]}

            # mix up
            # with torch.no_grad():
            input_labels = torch.cat([F.one_hot(y_lb, self.num_classes), sharpen_prob_x_ulb, sharpen_prob_x_ulb, sharpen_prob_x_ulb], dim=0)
            if self.mixup_manifold:
                inputs = torch.cat([outs_x_lb['feat'], outs_x_ulb_s_0['feat'], outs_x_ulb_s_1['feat'],  outs_x_ulb_w['feat']], dim=0)
            else:
                inputs = torch.cat([x_lb, x_ulb_s_0, x_ulb_s_1, x_ulb_w])
            mixed_x, mixed_y, _ = mixup_one_target(inputs, input_labels, self.mixup_alpha, is_bias=True)
            mixed_x = list(torch.split(mixed_x, num_lb))
            # mixed_x = interleave(mixed_x, num_lb)

            # calculate BN only for the first batch
            if self.mixup_manifold:
                logits = [self.model(mixed_x[0], only_fc=self.mixup_manifold)]
                # calculate BN for only the first batch
                self.bn_controller.freeze_bn(self.model)
                for ipt in mixed_x[1:]:
                    logits.append(self.model(ipt, only_fc=self.mixup_manifold))
                self.bn_controller.unfreeze_bn(self.model)
            else:
                logits = [self.model(mixed_x[0])['logits']]
                # calculate BN for only the first batch
                self.bn_controller.freeze_bn(self.model)
                for ipt in mixed_x[1:]:
                    logits.append(self.model(ipt)['logits'])
                self.bn_controller.unfreeze_bn(self.model)
            u1_logits = outs_x_ulb_s_0['logits']

            # put interleaved samples back
            # logits = interleave(logits, num_lb)
            logits_x = logits[0]
            logits_u = torch.cat(logits[1:], dim=0)

            # sup loss
            sup_loss = self.ce_loss(logits_x, mixed_y[:num_lb], reduction='mean')
            
            # unsup_loss
            unsup_loss = self.consistency_loss(logits_u, mixed_y[num_lb:])
            
            # loss U1
            u1_loss = self.consistency_loss(u1_logits, sharpen_prob_x_ulb)

            # ramp for w_match
            unsup_warmup = np.clip(self.it / (self.unsup_warm_up * self.num_train_iter),  a_min=0.0, a_max=1.0)
            total_loss = sup_loss + self.lambda_kl * unsup_warmup * u1_loss + self.lambda_u * unsup_warmup * unsup_loss

            # calculate rot loss with w_rot
            if self.use_rot:
                self.bn_controller.freeze_bn(self.model)
                logits_rot = self.model(x_ulb_s_0_rot, use_rot=True)['logits_rot']
                self.bn_controller.unfreeze_bn(self.model)
                rot_loss = F.cross_entropy(logits_rot, rot_v, reduction='mean')
                rot_loss = rot_loss.mean()
                total_loss += self.lambda_rot * rot_loss

        out_dict = self.process_out_dict(loss=total_loss, feat=feat_dict)
        log_dict = self.process_log_dict(sup_loss=sup_loss.item(), 
                                         unsup_loss=unsup_loss.item(), 
                                         total_loss=total_loss.item())
        return out_dict, log_dict


    def get_save_dict(self):
        save_dict = super().get_save_dict()
        save_dict['p_model'] = self.hooks_dict['DistAlignHook'].p_model.cpu()
        save_dict['p_target'] = self.hooks_dict['DistAlignHook'].p_target.cpu()
        return save_dict
    
    def load_model(self, load_path):
        checkpoint =  super().load_model(load_path)
        self.hooks_dict['DistAlignHook'].p_model = checkpoint['p_model'].cuda(self.args.gpu)
        self.hooks_dict['DistAlignHook'].p_target = checkpoint['p_target'].cuda(self.args.gpu)
        return checkpoint

    @staticmethod
    def get_argument():
        return [
            SSL_Argument('--T', float, 0.5, 'Temperature Sharpening'),
            SSL_Argument('--kl_loss_ratio', float, 0.5, 'weight for KL loss'),
            SSL_Argument('--rot_loss_ratio', float, 0.5, 'weight for rot loss, set to 0 for nlp and speech'),
            SSL_Argument('--unsup_warm_up', float, 1 / 64),
            SSL_Argument('--mixup_alpha', float, 0.75, 'param for Beta distribution of Mix Up'),
            SSL_Argument('--mixup_manifold', str2bool, False, 'use manifold mixup (for nlp)'),
        ]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import numpy as np
from semilearn.core import AlgorithmBase
from semilearn.core.utils import ALGORITHMS
from semilearn.algorithms.hooks import PseudoLabelingHook, FixedThresholdingHook
from semilearn.algorithms.utils import SSL_Argument


@ALGORITHMS.register('pseudolabel')
class PseudoLabel(AlgorithmBase):
    """
        Pseudo Label algorithm (https://arxiv.org/abs/1908.02983).

        Args:
        - args (`argparse`):
            algorithm arguments
        - net_builder (`callable`):
            network loading function
        - tb_log (`TBLog`):
            tensorboard logger
        - logger (`logging.Logger`):
            logger to use
        - p_cutoff(`float`):
            Confidence threshold for generating pseudo-labels
        - unsup_warm_up (`float`, *optional*, defaults to 0.4):
            Ramp up for weights for unsupervised loss
    """

    def __init__(self, args, net_builder, tb_log=None, logger=None, **kwargs):
        super().__init__(args, net_builder, tb_log, logger, **kwargs)
        self.init(p_cutoff=args.p_cutoff, unsup_warm_up=args.unsup_warm_up)

    def init(self, p_cutoff, unsup_warm_up=0.4):
        self.p_cutoff = p_cutoff
        self.unsup_warm_up = unsup_warm_up 

    def set_hooks(self):
        self.register_hook(PseudoLabelingHook(), "PseudoLabelingHook")
        self.register_hook(FixedThresholdingHook(), "MaskingHook")
        super().set_hooks()

    def train_step(self, x_lb, y_lb, x_ulb_w):
        # inference and calculate sup/unsup losses
        with self.amp_cm():

            outs_x_lb = self.model(x_lb)
            logits_x_lb = outs_x_lb['logits']
            feats_x_lb = outs_x_lb['feat']

            # calculate BN only for the first batch
            self.bn_controller.freeze_bn(self.model)
            outs_x_ulb = self.model(x_ulb_w)
            logits_x_ulb = outs_x_ulb['logits']
            feats_x_ulb = outs_x_ulb['feat']
            self.bn_controller.unfreeze_bn(self.model)

            feat_dict = {'x_lb': feats_x_lb, 'x_ulb_w': feats_x_ulb}

            sup_loss = self.ce_loss(logits_x_lb, y_lb, reduction='mean')

            # compute mask
            mask = self.call_hook("masking", "MaskingHook", logits_x_ulb=logits_x_ulb)

            # generate unlabeled targets using pseudo label hook
            pseudo_label = self.call_hook("gen_ulb_targets", "PseudoLabelingHook", 
                                          logits=logits_x_ulb,
                                          use_hard_label=True)

            unsup_loss = self.consistency_loss(logits_x_ulb,
                                               pseudo_label,
                                               'ce',
                                               mask=mask)

            unsup_warmup = np.clip(self.it / (self.unsup_warm_up * self.num_train_iter),  a_min=0.0, a_max=1.0)
            total_loss = sup_loss + self.lambda_u * unsup_loss * unsup_warmup

        out_dict = self.process_out_dict(loss=total_loss, feat=feat_dict)
        log_dict = self.process_log_dict(sup_loss=sup_loss.item(), 
                                         unsup_loss=unsup_loss.item(), 
                                         total_loss=total_loss.item(), 
                                         util_ratio=mask.float().mean().item())
        return out_dict, log_dict

    @staticmethod
    def get_argument():
        return [
            SSL_Argument('--p_cutoff', float, 0.95),
            SSL_Argument('--unsup_warm_up', float, 0.4, 'warm up ratio for unsupervised loss'),
            # SSL_Argument('--use_flex', str2bool, False),
        ]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch
import numpy as np
from semilearn.core.hooks import Hook


class MaskingHook(Hook):
    """
    Base MaskingHook, used for computing the mask of unlabeled (consistency) loss
    define MaskingHook in each algorithm when needed, and call hook inside each train_step
    easy support for other settings
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
    
    def update(self, *args, **kwargs):
        pass
    
    @torch.no_grad()
    def masking(self, algorithm, 
                      logits_x_lb=None, logits_x_ulb=None, idx_lb=None, idx_ulb=None,
                      softmax_x_lb=True, softmax_x_ulb=True,
                      *args, **kwargs):
        """
        generate mask for unlabeled loss

        Args:
            algorithm: base algorithm
            logits_x_lb: labeled batch logits (or probs, need to set softmax_x_lb to False)
            logits_x_ulb: unlabeled batch logits (or probs, need to set softmax_x_ulb to False)
            idx_lb: labeled batch index
            idx_ulb: unlabeled batch index
            softmax_x_lb: flag of using softmax on labeled logits 
            softmax_x_ulb: flag of using softmax on unlabeled logits 
        """
        raise NotImplementedError


class FixedThresholdingHook(MaskingHook):
    """
    Common Fixed Threshold used in fixmatch, uda, pseudo label, et. al.
    """

    @torch.no_grad()
    def masking(self, algorithm, logits_x_ulb, softmax_x_ulb=True, *args, **kwargs):
        if softmax_x_ulb:
            # probs_x_ulb = torch.softmax(logits_x_ulb.detach(), dim=-1)
            probs_x_ulb = algorithm.compute_prob(logits_x_ulb.detach())
        else:
            # logits is already probs
            probs_x_ulb = logits_x_ulb.detach()
        max_probs, _ = torch.max(probs_x_ulb, dim=-1)
        mask = max_probs.ge(algorithm.p_cutoff).to(max_probs.dtype)
        return mask


# class RampupWeightingHook(MaskingHook):
#     """
#     Common Rampup weight used in mean teacher, pi model, et. al.
#     """
#     def masking(self, algorithm, *args, **kwargs):
#         return np.clip(algorithm.it / (algorithm.unsup_warm_up * algorithm.num_train_iter),  a_min=0.0, a_max=1.0)


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
import torch
import torch.nn as nn
import ruamel.yaml as yaml
from torch.utils.tensorboard import SummaryWriter

def over_write_args_from_dict(args, dict):
    """
    overwrite arguments acocrding to config file
    """
    for k in dict:
        setattr(args, k, dict[k])


def over_write_args_from_file(args, yml):
    """
    overwrite arguments according to config file
    """
    if yml == '':
        return
    with open(yml, 'r', encoding='utf-8') as f:
        dic = yaml.load(f.read(), Loader=yaml.Loader)
        for k in dic:
            setattr(args, k, dic[k])


def setattr_cls_from_kwargs(cls, kwargs):
    # if default values are in the cls,
    # overlap the value by kwargs
    for key in kwargs.keys():
        if hasattr(cls, key):
            print(f"{key} in {cls} is overlapped by kwargs: {getattr(cls, key)} -> {kwargs[key]}")
        setattr(cls, key, kwargs[key])


def send_model_cuda(args, model, clip_batch=True):
    if not torch.cuda.is_available():
        raise Exception('ONLY GPU TRAINING IS SUPPORTED')
    elif args.distributed:
        ngpus_per_node = torch.cuda.device_count()  # number of gpus of each node

        if args.gpu is not None:
            torch.cuda.set_device(args.gpu)

            '''
            batch_size: batch_size per node -> batch_size per gpu
            workers: workers per node -> workers per gpu
            '''
            if clip_batch:
                args.batch_size = int(args.batch_size / ngpus_per_node)
            model.cuda(args.gpu)
            model = nn.SyncBatchNorm.convert_sync_batchnorm(model)
            model = torch.nn.parallel.DistributedDataParallel(model, broadcast_buffers=False,
                                                                     find_unused_parameters=True,
                                                                     device_ids=[args.gpu])
        else:
            # if arg.gpu is None, DDP will divide and allocate batch_size
            # to all available GPUs if device_ids are not set.
            model.cuda()
            model = torch.nn.parallel.DistributedDataParallel(model,  broadcast_buffers=False, 
                                                                      find_unused_parameters=True)
    elif args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        model = model.cuda(args.gpu)
    else:
        model = torch.nn.DataParallel(model).cuda()
    return model


def count_parameters(model):
    # count trainable parameters
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


class TBLog:
    """
    Construct tensorboard writer (self.writer).
    The tensorboard is saved at os.path.join(tb_dir, file_name).
    """

    def __init__(self, tb_dir, file_name, use_tensorboard=False):
        self.tb_dir = tb_dir
        self.use_tensorboard = use_tensorboard
        if self.use_tensorboard:
            self.writer = SummaryWriter(os.path.join(self.tb_dir, file_name))
            

    def update(self, log_dict, it, suffix=None, mode="train"):
        """
        Args
            log_dict: contains scalar values for updating tensorboard
            it: contains information of iteration (int).
            suffix: If not None, the update key has the suffix.
        """
        if suffix is None:
            suffix = ''
        if self.use_tensorboard:
            for key, value in log_dict.items():
                self.writer.add_scalar(suffix + key, value, it)


class Bn_Controller:
    """
    Batch Norm controller
    """
    def __init__(self):
        """
        freeze_bn and unfreeze_bn must appear in pairs
        """
        self.backup = {}

    def freeze_bn(self, model):
        assert self.backup == {}
        for name, m in model.named_modules():
            if isinstance(m, nn.SyncBatchNorm) or isinstance(m, nn.BatchNorm2d):
                self.backup[name + '.running_mean'] = m.running_mean.data.clone()
                self.backup[name + '.running_var'] = m.running_var.data.clone()
                self.backup[name + '.num_batches_tracked'] = m.num_batches_tracked.data.clone()

    def unfreeze_bn(self, model):
        for name, m in model.named_modules():
            if isinstance(m, nn.SyncBatchNorm) or isinstance(m, nn.BatchNorm2d):
                m.running_mean.data = self.backup[name + '.running_mean']
                m.running_var.data = self.backup[name + '.running_var']
                m.num_batches_tracked.data = self.backup[name + '.num_batches_tracked']
        self.backup = {}


class EMA:
    """
    EMA model
    Implementation from https://fyubang.com/2019/06/01/ema/
    """

    def __init__(self, model, decay):
        self.model = model
        self.decay = decay
        self.shadow = {}
        self.backup = {}

    def load(self, ema_model):
        for name, param in ema_model.named_parameters():
            self.shadow[name] = param.data.clone()

    def register(self):
        for name, param in self.model.named_parameters():
            self.shadow[name] = param.data.clone()

    def update(self):
        for name, param in self.model.named_parameters():
            new_average = (1.0 - self.decay) * param.data + self.decay * self.shadow[name]
            self.shadow[name] = new_average.clone()

    def apply_shadow(self):
        for name, param in self.model.named_parameters():
            self.backup[name] = param.data
            param.data = self.shadow[name]

    def restore(self):
        for name, param in self.model.named_parameters():
            param.data = self.backup[name]
        self.backup = {}



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch
import torch.nn as nn
import torch.nn.functional as F
from semilearn.core.criterions import CELoss


class TRASLogitsAdjCELoss(CELoss):
    def __init__(self, la):
        super().__init__()
        self.la = la

    def forward(self, logits, targets, reduction='mean'):
        return super().forward(logits + self.la, targets, reduction=reduction)

class TRASKLLoss(nn.Module):
    def forward(self, outputs, targets, T, mask):
        _p = F.log_softmax(outputs / T, dim=1)
        _q = F.softmax(targets / (T * 2), dim=1)
        _soft_loss = -torch.mean(torch.sum(_q * _p, dim=1) * mask)
        _soft_loss = _soft_loss * T * T
        return _soft_loss



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .core.utils import get_dataset, get_data_loader, get_net_builder
from .algorithms import get_algorithm
from .datasets import split_ssl_data
from .datasets.cv_datasets.datasetbase import BasicDataset
from .lighting import Trainer, get_config




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from semilearn.datasets.utils import split_ssl_data, get_collactor
from semilearn.datasets.cv_datasets import get_cifar, get_eurosat, get_imagenet, get_medmnist, get_semi_aves, get_stl10, get_svhn, get_food101
from semilearn.datasets.nlp_datasets import get_json_dset
from semilearn.datasets.audio_datasets import get_pkl_dset
from semilearn.datasets.samplers import name2sampler, DistributedSampler, WeightedDistributedSampler, ImageNetDistributedSampler



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch
from .utils import DebiasPLConsistencyLoss, DebiasPLHook

from semilearn.core import ImbAlgorithmBase
from semilearn.core.utils import IMB_ALGORITHMS
from semilearn.algorithms.utils import SSL_Argument


@IMB_ALGORITHMS.register('debiaspl')
class DebiasPL(ImbAlgorithmBase):
    """
        DebiasPL algorithm (https://openaccess.thecvf.com/content/CVPR2022/papers/Wang_Debiased_Learning_From_Naturally_Imbalanced_Pseudo-Labels_CVPR_2022_paper.pdf).

        Args:
            - args (`argparse`):
                algorithm arguments
            - net_builder (`callable`):
                network loading function
            - tb_log (`TBLog`):
                tensorboard logger
            - logger (`logging.Logger`):
                logger to use
            - debiaspl_tau (float):
                tau in DebiasPl
            - debiaspl_ema_p (float):
                momentum 
    """
    def __init__(self, args, **kwargs):
        self.imb_init(args.debiaspl_tau, args.debiaspl_ema_p)
        super().__init__(args, **kwargs)
        assert args.algorithm not in ['mixmatch', 'meanteacher', 'pimodel'], "DebiasPL not supports {} as the base algorithm.".format(args.algorithm)

        self.p_hat = torch.ones((self.num_classes, )).to(self.gpu) / self.num_classes
        self.consistency_loss = DebiasPLConsistencyLoss(tau=self.tau)


    def imb_init(self, tau=0.4, ema_p=0.999):
        self.tau = tau 
        self.ema_p = ema_p

    def set_hooks(self):
        super().set_hooks()
        self.register_hook(DebiasPLHook(), "NORMAL")


    def compute_prob(self, logits):
        # update p_hat
        probs = super().compute_prob(logits)
        delta_p = probs.mean(dim=0)
        self.p_hat = self.ema_m * self.p_hat + (1 - self.ema_p) * delta_p
        return super().compute_prob(logits - self.tau * torch.log(self.p_hat))

    @staticmethod
    def get_argument():
        return [
            SSL_Argument('--debiaspl_tau', float, 0.4),
            SSL_Argument('--debiaspl_ema_p', float, 0.999),
        ]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import copy
from inspect import signature
import torch
import numpy as np


from .utils import SAWWeightsHook, SAWCELoss, SAWConsistencyLoss

from semilearn.core import ImbAlgorithmBase
from semilearn.core.utils import get_data_loader, IMB_ALGORITHMS


@IMB_ALGORITHMS.register('saw')
class SAW(ImbAlgorithmBase):
    """
        SAW algorithm (https://proceedings.mlr.press/v162/lai22b.html).

        Args:
            - args (`argparse`):
                algorithm arguments
            - net_builder (`callable`):
                network loading function
            - tb_log (`TBLog`):
                tensorboard logger
            - logger (`logging.Logger`):
                logger to use
    """
    def __init__(self, args, net_builder, tb_log=None, logger=None, **kwargs):
        super().__init__(args, net_builder, tb_log, logger, **kwargs)
        self.ce_loss = SAWCELoss()
        self.consistency_loss = SAWConsistencyLoss()


    def set_hooks(self):
        super().set_hooks()

        # get ground truth distribution
        y_lb_cnt = [0 for _ in range(self.num_classes)]
        for c in self.dataset_dict['train_lb'].targets:
            y_lb_cnt[c] += 1
        y_lb_cnt = torch.from_numpy(np.array(y_lb_cnt))
        num_ulb = len(self.dataset_dict['train_ulb'])

        # add weight hooks
        self.register_hook(SAWWeightsHook(y_lb_cnt=y_lb_cnt, num_ulb=num_ulb),
                           "SAWWeightsHook", "NORMAL")


    def set_dataset(self):
        dataset_dict = super().set_dataset()
        dataset_dict['eval_ulb'] = copy.deepcopy(dataset_dict['train_ulb'])
        dataset_dict['eval_ulb'].is_ulb = False
        return dataset_dict
    
    def set_data_loader(self):
        loader_dict = super().set_data_loader()
        
        # add unlabeled evaluation data loader
        loader_dict['eval_ulb'] = get_data_loader(self.args,
                                                  self.dataset_dict['eval_ulb'],
                                                  self.args.eval_batch_size,
                                                  data_sampler=None,
                                                  shuffle=False,
                                                  num_workers=self.args.num_workers,
                                                  drop_last=False)
        return loader_dict


    def get_save_dict(self):
        save_dict = super().get_save_dict()
        save_dict['x_lb_weights'] = self.x_lb_weights.cpu()
        save_dict['x_ulb_weights'] = self.x_ulb_weights.cpu()
        return save_dict


    def load_model(self, load_path):
        checkpoint = super().load_model(load_path)
        self.x_lb_weights = checkpoint['x_lb_weights'].cuda(self.gpu)
        self.x_ulb_weights = checkpoint['x_ulb_weights'].cuda(self.gpu)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
import torch
import torch.nn as nn
import numpy as np
from inspect import signature

from semilearn.core import ImbAlgorithmBase
from semilearn.core.utils import IMB_ALGORITHMS
from semilearn.algorithms.utils import SSL_Argument


class ABCNet(nn.Module):
    def __init__(self, backbone, num_classes):
        super().__init__()
        self.backbone = backbone
        self.num_features = backbone.num_features

        # auxiliary classifier
        self.aux_classifier = nn.Linear(self.backbone.num_features, num_classes)
    
    def forward(self, x, **kwargs):
        results_dict = self.backbone(x, **kwargs)
        results_dict['logits_aux'] = self.aux_classifier(results_dict['feat'])
        return results_dict

    def group_matcher(self, coarse=False):
        if hasattr(self.backbone, 'backbone'):
            # TODO: better way
            matcher = self.backbone.backbone.group_matcher(coarse, prefix='backbone.backbone')
        else:
            matcher = self.backbone.group_matcher(coarse, prefix='backbone.')
        return matcher


@IMB_ALGORITHMS.register('abc')
class ABC(ImbAlgorithmBase):
    """
        ABC algorithm (https://arxiv.org/abs/2110.10368).

        Args:
            - args (`argparse`):
                algorithm arguments
            - net_builder (`callable`):
                network loading function
            - tb_log (`TBLog`):
                tensorboard logger
            - logger (`logging.Logger`):
                logger to use
            - abc_p_cutoff (`float`):
                threshold for the auxilariy classifier
            - abc_loss_ratio (`float`):
                loss ration for auxiliary classifier
    """
    def __init__(self, args, net_builder, tb_log=None, logger=None, **kwargs):
        self.imb_init(abc_p_cutoff=args.abc_p_cutoff, abc_loss_ratio=args.abc_loss_ratio)

        super(ABC, self).__init__(args, net_builder, tb_log, logger, **kwargs)

        # compute lb imb ratio
        lb_class_dist = [0 for _ in range(self.num_classes)]
        for c in  self.dataset_dict['train_lb'].targets:
            lb_class_dist[c] += 1
        lb_class_dist = np.array(lb_class_dist)
        self.lb_class_dist = torch.from_numpy(np.min(lb_class_dist) / lb_class_dist)
        
        # TODO: better ways
        self.model = ABCNet(self.model, num_classes=self.num_classes)
        self.ema_model = ABCNet(self.ema_model, num_classes=self.num_classes)
        self.ema_model.load_state_dict(self.model.state_dict())
        self.optimizer, self.scheduler = self.set_optimizer()

    def imb_init(self, abc_p_cutoff=0.95, abc_loss_ratio=1.0):
        self.abc_p_cutoff = abc_p_cutoff
        self.abc_loss_ratio = abc_loss_ratio

    def process_batch(self, **kwargs):
        # get core algorithm parameters
        input_args = signature(super().train_step).parameters
        input_args = list(input_args.keys())
        return super().process_batch(input_args=input_args, **kwargs)

    def train_step(self, *args, **kwargs):
        
        out_dict, log_dict = super().train_step(*args, **kwargs)

        # get features
        feats_x_lb = out_dict['feat']['x_lb']
        feats_x_ulb_w = out_dict['feat']['x_ulb_w']
        feats_x_ulb_s = out_dict['feat']['x_ulb_s']
        if isinstance(feats_x_ulb_s, list):
            feats_x_ulb_s = feats_x_ulb_s[0]
        
        # get logits
        logits_x_lb = self.model.module.aux_classifier(feats_x_lb)
        logits_x_ulb_s = self.model.module.aux_classifier(feats_x_ulb_s)
        with torch.no_grad():
            logits_x_ulb_w = self.model.module.aux_classifier(feats_x_ulb_w)

        # compute abc loss using logits_aux from dict
        abc_loss = self.compute_abc_loss(
            logits_x_lb=logits_x_lb, 
            y_lb=kwargs['y_lb'],
            logits_x_ulb_w=logits_x_ulb_w,
            logits_x_ulb_s=logits_x_ulb_s
            )
        out_dict['loss'] += self.abc_loss_ratio * abc_loss 
        log_dict['train/abc_loss'] = abc_loss.item()

        return out_dict, log_dict
    
    def evaluate(self, eval_dest='eval', out_key='logits', return_logits=False):
        return super().evaluate(eval_dest=eval_dest, out_key='logits_aux', return_logits=return_logits)

    @staticmethod
    @torch.no_grad()
    def bernouli_mask(x):
        return torch.bernoulli(x.detach()).float()
    
    def compute_abc_loss(self, logits_x_lb, y_lb, logits_x_ulb_w, logits_x_ulb_s):
        if not isinstance(logits_x_ulb_s, list):
            logits_x_ulb_s = [logits_x_ulb_s]
        
        if not self.lb_class_dist.is_cuda:
            self.lb_class_dist = self.lb_class_dist.to(y_lb.device)

        # compute labeled abc loss
        mask_lb = self.bernouli_mask(self.lb_class_dist[y_lb])
        abc_lb_loss = (self.ce_loss(logits_x_lb, y_lb, reduction='none') * mask_lb).mean()

        # compute unlabeled abc loss
        with torch.no_grad():
            # probs_x_ulb_w = torch.softmax(logits_x_ulb_w, dim=1)
            probs_x_ulb_w = self.compute_prob(logits_x_ulb_w)
            max_probs, y_ulb = torch.max(probs_x_ulb_w, dim=1)
            mask_ulb_1 = max_probs.ge(self.abc_p_cutoff).to(logits_x_ulb_w.dtype)
            ulb_class_dist = 1 - (self.epoch / self.epochs) * (1 - self.lb_class_dist)
            mask_ulb_2 = self.bernouli_mask(ulb_class_dist[y_ulb])
            mask_ulb = mask_ulb_1 * mask_ulb_2
    
        abc_ulb_loss = 0.0
        for logits_s in logits_x_ulb_s:
            abc_ulb_loss += (self.ce_loss(logits_s, y_ulb, reduction='none') * mask_ulb).mean()
        
        abc_loss = abc_lb_loss + abc_ulb_loss
        return abc_loss


    @staticmethod
    def get_argument():
        return [
            SSL_Argument('--abc_p_cutoff', float, 0.95),
            SSL_Argument('--abc_loss_ratio', float, 1.0),
        ]        



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch
from copy import deepcopy
from collections import Counter

from semilearn.algorithms.hooks import MaskingHook


class FlexMatchThresholdingHook(MaskingHook):
    """
    Adaptive Thresholding in FlexMatch
    """
    def __init__(self, ulb_dest_len, num_classes, thresh_warmup=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ulb_dest_len = ulb_dest_len
        self.num_classes = num_classes
        self.thresh_warmup = thresh_warmup
        self.selected_label = torch.ones((self.ulb_dest_len,), dtype=torch.long, ) * -1
        self.classwise_acc = torch.zeros((self.num_classes,))

    @torch.no_grad()
    def update(self, *args, **kwargs):
        pseudo_counter = Counter(self.selected_label.tolist())
        if max(pseudo_counter.values()) < self.ulb_dest_len:  # not all(5w) -1
            if self.thresh_warmup:
                for i in range(self.num_classes):
                    self.classwise_acc[i] = pseudo_counter[i] / max(pseudo_counter.values())
            else:
                wo_negative_one = deepcopy(pseudo_counter)
                if -1 in wo_negative_one.keys():
                    wo_negative_one.pop(-1)
                for i in range(self.num_classes):
                    self.classwise_acc[i] = pseudo_counter[i] / max(wo_negative_one.values())

    @torch.no_grad()
    def masking(self, algorithm, logits_x_ulb, idx_ulb, softmax_x_ulb=True, *args, **kwargs):
        if not self.selected_label.is_cuda:
            self.selected_label = self.selected_label.to(logits_x_ulb.device)
        if not self.classwise_acc.is_cuda:
            self.classwise_acc = self.classwise_acc.to(logits_x_ulb.device)

        if softmax_x_ulb:
            # probs_x_ulb = torch.softmax(logits_x_ulb.detach(), dim=-1)
            probs_x_ulb = self.compute_prob(logits_x_ulb.detach())
        else:
            # logits is already probs
            probs_x_ulb = logits_x_ulb.detach()
        max_probs, max_idx = torch.max(probs_x_ulb, dim=-1)
        # mask = max_probs.ge(p_cutoff * (class_acc[max_idx] + 1.) / 2).float()  # linear
        # mask = max_probs.ge(p_cutoff * (1 / (2. - class_acc[max_idx]))).float()  # low_limit
        mask = max_probs.ge(algorithm.p_cutoff * (self.classwise_acc[max_idx] / (2. - self.classwise_acc[max_idx])))  # convex
        # mask = max_probs.ge(p_cutoff * (torch.log(class_acc[max_idx] + 1.) + 0.5)/(math.log(2) + 0.5)).float()  # concave
        select = max_probs.ge(algorithm.p_cutoff)
        mask = mask.to(max_probs.dtype)

        # update
        if idx_ulb[select == 1].nelement() != 0:
            self.selected_label[idx_ulb[select == 1]] = max_idx[select == 1]
        self.update()

        return mask
        
        





from __future__ import print_function
import copy
import time
import numpy as np

import torch
import torch.optim as optim
import torch.utils.data as data
import torch.nn.functional as F
from torch.utils.data.sampler import WeightedRandomSampler


from progress.bar import Bar as Bar

__all__ = ['classifier_warmup']


class AverageMeter(object):
    """Computes and stores the average and current value
       Imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class WeightEMA(object):
    def __init__(self, model, ema_model, lr, alpha=0.999, wd=True):
        self.model = model
        self.ema_model = ema_model
        self.alpha = alpha
        self.params = list(model.state_dict().values())
        self.ema_params = list(ema_model.state_dict().values())
        # wd is implemented here to be compatible with https://github.com/bbuing9/DARP
        if wd:
            self.wd = 0.02 * lr
        else:
            self.wd = 0.0

        for param, ema_param in zip(self.params, self.ema_params):
            param.data.copy_(ema_param.data)

    def step(self):
        one_minus_alpha = 1.0 - self.alpha
        for param, ema_param in zip(self.params, self.ema_params):
            ema_param = ema_param.float()
            param = param.float()
            ema_param.mul_(self.alpha)
            ema_param.add_(param * one_minus_alpha)
            # customized weight decay
            param.mul_(1 - self.wd)


def classifier_warmup(args, model, train_labeled_set, train_unlabeled_set, N_SAMPLES_PER_CLASS, num_class, gpu):

    # Hypers used during warmup
    epochs = args.cossl_tfe_warm_epoch  # 10
    lr = args.cossl_tfe_warm_lr  # 0.002
    ema_decay = args.cossl_tfe_warm_ema_decay  # 0.999
    weight_decay = args.cossl_tfe_warm_wd  # 5e-4
    batch_size = args.cossl_tfe_warm_bs  # 64
    val_iteration = args.num_eval_iter  # 500

    # Construct dataloaders
    labeled_trainloader = data.DataLoader(train_labeled_set, batch_size=batch_size,
                                          shuffle=False, num_workers=0, drop_last=False)
    unlabeled_trainloader = data.DataLoader(train_unlabeled_set, batch_size=batch_size,
                                            shuffle=False, num_workers=0, drop_last=False)

    # weight_imprint is not necessary
    tfe_model = weight_imprint(args, copy.deepcopy(model), train_labeled_set, num_class, gpu)

    # Fix the feature extractor and reinitialize the classifier
    for param in model.parameters():
        param.requires_grad = False
    
    model.module.classifier.reset_parameters()
    for param in model.module.classifier.parameters():
        param.requires_grad = True

    model = model.cuda(gpu)
    ema_model = copy.deepcopy(model)
    for param in ema_model.parameters():
        param.detach_()
    ema_optimizer = WeightEMA(model, ema_model, lr, alpha=ema_decay, wd=False)

    wd_params, non_wd_params = [], []
    for name, param in model.module.classifier.named_parameters():
        if 'bn' in name or 'bias' in name:
            non_wd_params.append(param)  # bn.weight, bn.bias and classifier.bias, conv2d.bias
        else:
            wd_params.append(param)
    param_list = [{'params': wd_params, 'weight_decay': weight_decay}, {'params': non_wd_params, 'weight_decay': 0}]
    print('    Total params: %.2fM' % (sum(p.numel() for p in model.module.classifier.parameters()) / 1000000.0))

    optimizer = optim.Adam(param_list, lr=1e-3)

    # Generate TFE features in advance as the model and the data loaders are fixed anyway
    balanced_feature_set = TFE(args, labeled_trainloader, unlabeled_trainloader,
                               tfe_model, num_class, N_SAMPLES_PER_CLASS, gpu)
    balanced_feature_loader = data.DataLoader(balanced_feature_set, batch_size=batch_size,
                                              shuffle=True, num_workers=0, drop_last=True)

    # Main function
    for epoch in range(epochs):
        print('\nTFE-head-warming: Epoch: [%d | %d] LR: %f' % (epoch + 1, epochs, optimizer.param_groups[0]['lr']))

        classifier_train(args, balanced_feature_loader, model, optimizer, None, ema_optimizer, val_iteration, gpu)

    return model, ema_model


def TFE(args, labeled_loader, unlabeled_loader, tfe_model, num_classes, num_samples_per_class, gpu):

    tfe_model.eval()
    with torch.no_grad():
        # ****************** extract features  ********************
        # extract features from labeled data
        for batch_idx, data_dict in enumerate(labeled_loader):
            inputs = data_dict['x_lb']
            targets = data_dict['y_lb']

            inputs = inputs.cuda(gpu)
            targets = targets.cuda(gpu)
            ret_dict = tfe_model(inputs)
            logits = ret_dict['logits']
            features = ret_dict['feat']
            cls_probs = torch.softmax(logits, dim=1)
            features = features.squeeze()  # Note: a flatten is needed here
            if batch_idx == 0:
                labeled_feature_stack = features
                labeled_target_stack = targets
                labeled_cls_prob_stack = cls_probs
            else:
                labeled_feature_stack = torch.cat((labeled_feature_stack, features), 0)
                labeled_target_stack = torch.cat((labeled_target_stack, targets), 0)
                labeled_cls_prob_stack = torch.cat((labeled_cls_prob_stack, cls_probs), 0)
        # extract features from unlabeled data
        for batch_idx, data_dict in enumerate(unlabeled_loader):
            inputs_w = data_dict['x_ulb_w'].cuda(gpu)
            if args.algorithm in ['remixmatch', 'comatch']:
                inputs_s = data_dict['x_ulb_s_0'].cuda(gpu)
            else:
                inputs_s = data_dict['x_ulb_s'].cuda(gpu)
            features = tfe_model(inputs_s)['feat']
            logits = tfe_model(inputs_w)['logits']
            # if hasattr(unlabeled_loader.dataset.transform, 'transform2'):  # FixMatch, ReMixMatch
            #     inputs_w, inputs_s, _ = data_batch
            #     inputs_s = inputs_s.cuda()
            #     inputs_w = inputs_w.cuda()
            #
            #     features = tfe_model(inputs_s)['feat']
            #     logits = tfe_model(inputs_w)['logits']
            # else:  # MixMatch
            #     inputs_w, _ = data_batch
            #     inputs_w = inputs_w.cuda()
            #     ret_dict = tfe_model(inputs_w)
            #     logits = ret_dict['logits']
            #     features = ret_dict['feat']
            cls_probs = torch.softmax(logits, dim=1)
            _, targets = torch.max(cls_probs, dim=1)
            features = features.squeeze()
            if batch_idx == 0:
                unlabeled_feature_stack = features
                unlabeled_target_stack = targets
                unlabeled_cls_prob_stack = cls_probs
            else:
                unlabeled_feature_stack = torch.cat((unlabeled_feature_stack, features), 0)
                unlabeled_target_stack = torch.cat((unlabeled_target_stack, targets), 0)
                unlabeled_cls_prob_stack = torch.cat((unlabeled_cls_prob_stack, cls_probs), 0)

        # ****************** create TFE features for each class  ********************
        # create idx array for each class, per_cls_idx[i] contains all indices of images of class i
        labeled_set_idx = torch.tensor(list(range(len(labeled_feature_stack))))
        labeled_set_per_cls_idx = [labeled_set_idx[(labeled_target_stack == i).to(labeled_set_idx.device)] for i in range(num_classes)]

        augment_features = []  # newly generated tfe features will be appended here
        augment_targets = []  # as well as their one-hot targets
        for cls_id in range(num_classes):
            if num_samples_per_class[cls_id] < max(num_samples_per_class):

                # how many we need for the cls
                augment_size = max(num_samples_per_class) - num_samples_per_class[cls_id]

                # create data belonging to class i
                current_cls_feats = labeled_feature_stack[labeled_target_stack == cls_id]

                # create data not belonging to class i
                other_labeled_data_idx = np.concatenate(labeled_set_per_cls_idx[:cls_id] + labeled_set_per_cls_idx[cls_id + 1:], axis=0)
                other_cls_feats = torch.cat([labeled_feature_stack[other_labeled_data_idx], unlabeled_feature_stack], dim=0)
                other_cls_probs = torch.cat([labeled_cls_prob_stack[other_labeled_data_idx], unlabeled_cls_prob_stack], dim=0)

                assert len(other_cls_feats) == len(other_cls_probs)
                # the total number of data should be the same for label-unlabel split, and current-the-rest split
                assert (len(other_cls_feats) + len(current_cls_feats)) == (len(labeled_feature_stack) + len(unlabeled_feature_stack))

                # sort other_cls_feats according to the probs assigned to class i
                probs4current_cls = other_cls_probs[:, cls_id]
                sorted_probs, order = probs4current_cls.sort(descending=True)  # sorted_probs = probs belonging to cls i
                other_cls_feats = other_cls_feats[order]

                # select features from the current class
                input_a_idx = np.random.choice(list(range(len(current_cls_feats))), augment_size, replace=True)
                # take first n features from all other classes
                input_b_idx = np.asarray(list(range(augment_size)))
                lambdas = np.random.beta(0.75, 0.75, size=augment_size)

                # do TFE
                for l, a_idx, b_idx in zip(lambdas, input_a_idx, input_b_idx):
                    tfe_input = l * current_cls_feats[a_idx] + (1 - l) * other_cls_feats[b_idx]  # [128]
                    tfe_target = torch.zeros((1, num_classes))
                    tfe_target[0, cls_id] = 1  # pseudo_label.tolist()
                    augment_features.append(tfe_input.view(1, -1))
                    augment_targets.append(tfe_target)

        # ****************** merge newly generated data with labeled dataset  ********************
        augment_features = torch.cat(augment_features, dim=0)
        augment_targets = torch.cat(augment_targets, dim=0).cuda(gpu)

        target_stack = torch.zeros(len(labeled_target_stack), num_classes).cuda(gpu).scatter_(1, labeled_target_stack.view(-1, 1), 1)
        new_feat_tensor = torch.cat([labeled_feature_stack, augment_features], dim=0)
        new_target_tensor = torch.cat([target_stack, augment_targets], dim=0)

    balanced_feature_set = data.dataset.TensorDataset(new_feat_tensor, new_target_tensor)
    return balanced_feature_set


def weight_imprint(args, model, labeled_set, num_classes, gpu):
    model = model.cuda(gpu)
    model.eval()

    labeledloader = data.DataLoader(labeled_set, batch_size=100, shuffle=False, num_workers=0, drop_last=False)

    with torch.no_grad():
        bar = Bar('Classifier weight imprinting...', max=len(labeledloader))

        for batch_idx, data_dict in enumerate(labeledloader):
            inputs = data_dict['x_lb']
            targets = data_dict['y_lb']

            inputs = inputs.cuda(gpu)
            features = model(inputs)['feat']
            output = features.squeeze()   # Note: a flatten is needed here

            if batch_idx == 0:
                output_stack = output
                target_stack = targets
            else:
                output_stack = torch.cat((output_stack, output), 0)
                target_stack = torch.cat((target_stack, targets), 0)

            bar.suffix = '({batch}/{size}'.format(batch=batch_idx + 1, size=len(labeledloader))
            bar.next()
        bar.finish()

    
    new_weight = torch.zeros(num_classes, model.module.backbone.num_features)

    for i in range(num_classes):
        tmp = output_stack[target_stack == i].mean(0)
        new_weight[i] = tmp / tmp.norm(p=2)
    model.module.classifier = torch.nn.Linear(model.module.backbone.num_features, num_classes, bias=False).cuda(gpu)
        # model.classifier.reset_parameters()
    model.module.classifier.weight.data = new_weight.cuda(gpu)

    model.eval()
    return model


def classifier_train(args, labeled_trainloader, model, optimizer, scheduler, ema_optimizer, val_iteration, gpu):
    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    train_acc = AverageMeter()
    end = time.time()

    # bar = Bar('Training', max=val_iteration)
    labeled_train_iter = iter(labeled_trainloader)

    model.eval()
    for batch_idx in range(val_iteration):
        try:
            inputs_x, targets_x = next(labeled_train_iter)
        except:
            labeled_train_iter = iter(labeled_trainloader)
            inputs_x, targets_x = next(labeled_train_iter)

        data_time.update(time.time() - end)

        inputs_x, targets_x = inputs_x.cuda(gpu), targets_x.cuda(gpu)  # targets are one-hot

        outputs = model.module.classifier(inputs_x)

        loss = (-F.log_softmax(outputs, dim=1) * targets_x).sum(dim=1)
        loss = loss.mean()
        acc = (torch.argmax(outputs, dim=1) == torch.argmax(targets_x, dim=1)).float().sum() / len(targets_x)

        # Record loss and acc
        losses.update(loss.item(), inputs_x.size(0))
        train_acc.update(acc.item(), inputs_x.size(0))

        # Compute gradient and apply SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        ema_optimizer.step()
        if scheduler is not None:
            scheduler.step()

        batch_time.update(time.time() - end)
        end = time.time()

    return losses.avg, train_acc.avg


def get_weighted_sampler(target_sample_rate, num_sample_per_class, target):
    assert len(num_sample_per_class) == len(np.unique(target))

    sample_weights = target_sample_rate / num_sample_per_class  # this is the key line!!!
    print(sample_weights)

    # assign each sample a weight by sampling rate
    samples_weight = np.array([sample_weights[t] for t in target])

    return WeightedRandomSampler(samples_weight, len(samples_weight), True)


def make_imb_data(max_num, class_num, gamma):
    mu = np.power(1/abs(gamma), 1/(class_num - 1))
    class_num_list = []
    for i in range(class_num):
        if i == (class_num - 1):
            class_num_list.append(int(max_num / abs(gamma)))
        else:
            class_num_list.append(int(max_num * np.power(mu, i)))
    if gamma < 0:
        class_num_list = class_num_list[::-1]
    return list(class_num_list)


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# Ref: https://github.com/open-mmlab/mmcv/blob/master/mmcv/runner/hooks/checkpoint.py

import os

from .hook import Hook

class CheckpointHook(Hook):
    """
    Checkpoint Hook for saving checkpoint
    """
    def after_train_step(self, algorithm):
        # must be called after evaluation for saving the best
        if self.every_n_iters(algorithm, algorithm.num_eval_iter) or self.is_last_iter(algorithm):
            save_path = os.path.join(algorithm.save_dir, algorithm.save_name)
            
            if (not algorithm.distributed) or \
               (algorithm.distributed and algorithm.rank % algorithm.ngpus_per_node == 0):
                algorithm.save_model('latest_model.pth', save_path)

                if algorithm.it == algorithm.best_it:
                    algorithm.save_model('model_best.pth', save_path)

        